from django.core.exceptions import ValidationError
from django.db import IntegrityError
from django.test import TestCase

from .models import UniqueIntegers


class SmallIntegerFieldTests(TestCase):
    max_value = 2**31 - 1
    min_value = -(2**31)

    def test_unique_max_value(self):
        """
        SmallIntegerField.db_type() is "int" which means unique constraints
        are only enforced up to 32-bit values.
        """
        UniqueIntegers.objects.create(small=self.max_value + 1)
        UniqueIntegers.objects.create(small=self.max_value + 1)  # no IntegrityError
        UniqueIntegers.objects.create(small=self.max_value)
        with self.assertRaises(IntegrityError):
            UniqueIntegers.objects.create(small=self.max_value)

    def test_unique_min_value(self):
        """
        SmallIntegerField.db_type() is "int" which means unique constraints
        are only enforced down to negative 32-bit values.
        """
        UniqueIntegers.objects.create(small=self.min_value - 1)
        UniqueIntegers.objects.create(small=self.min_value - 1)  # no IntegrityError
        UniqueIntegers.objects.create(small=self.min_value)
        with self.assertRaises(IntegrityError):
            UniqueIntegers.objects.create(small=self.min_value)

    def test_validate_max_value(self):
        UniqueIntegers(small=self.max_value).full_clean()  # no error
        msg = "{'small': ['Ensure this value is less than or equal to 2147483647.']"
        with self.assertRaisesMessage(ValidationError, msg):
            UniqueIntegers(small=self.max_value + 1).full_clean()

    def test_validate_min_value(self):
        UniqueIntegers(small=self.min_value).full_clean()  # no error
        msg = "{'small': ['Ensure this value is greater than or equal to -2147483648.']"
        with self.assertRaisesMessage(ValidationError, msg):
            UniqueIntegers(small=self.min_value - 1).full_clean()


class PositiveSmallIntegerFieldTests(TestCase):
    max_value = 2**31 - 1
    min_value = 0

    def test_unique_max_value(self):
        """
        SmallIntegerField.db_type() is "int" which means unique constraints
        are only enforced up to 32-bit values.
        """
        UniqueIntegers.objects.create(positive_small=self.max_value + 1)
        UniqueIntegers.objects.create(positive_small=self.max_value + 1)  # no IntegrityError
        UniqueIntegers.objects.create(positive_small=self.max_value)
        with self.assertRaises(IntegrityError):
            UniqueIntegers.objects.create(positive_small=self.max_value)

    # test_unique_min_value isn't needed since PositiveSmallIntegerField has a
    # limit of zero (enforced only in forms and model validation).

    def test_validate_max_value(self):
        UniqueIntegers(positive_small=self.max_value).full_clean()  # no error
        msg = "{'positive_small': ['Ensure this value is less than or equal to 2147483647.']"
        with self.assertRaisesMessage(ValidationError, msg):
            UniqueIntegers(positive_small=self.max_value + 1).full_clean()

    def test_validate_min_value(self):
        UniqueIntegers(positive_small=self.min_value).full_clean()  # no error
        msg = "{'positive_small': ['Ensure this value is greater than or equal to 0.']"
        with self.assertRaisesMessage(ValidationError, msg):
            UniqueIntegers(positive_small=self.min_value - 1).full_clean()


class SmallUniqueTests(TestCase):
    """
    Duplicate values < 32 bits are prohibited. This confirms integer field
    values are cast to Int64 so MongoDB stores it as long. Otherwise, the
    partialFilterExpression: {$type: long} unique constraint doesn't work.
    """

    test_value = 123

    def test_integerfield(self):
        UniqueIntegers.objects.create(plain=self.test_value)
        with self.assertRaisesMessage(IntegrityError, "duplicate key error"):
            UniqueIntegers.objects.create(plain=self.test_value)

    def test_bigintegerfield(self):
        UniqueIntegers.objects.create(big=self.test_value)
        with self.assertRaises(IntegrityError):
            UniqueIntegers.objects.create(big=self.test_value)

    def test_positiveintegerfield(self):
        UniqueIntegers.objects.create(positive=self.test_value)
        with self.assertRaises(IntegrityError):
            UniqueIntegers.objects.create(positive=self.test_value)

    def test_positivebigintegerfield(self):
        UniqueIntegers.objects.create(positive_big=self.test_value)
        with self.assertRaises(IntegrityError):
            UniqueIntegers.objects.create(positive_big=self.test_value)


class LargeUniqueTests(TestCase):
    """
    Duplicate values > 32 bits are prohibited. This confirms each field uses
    the long db_type() rather than the 32 bit int type.
    """

    test_value = 2**63 - 1

    def test_integerfield(self):
        UniqueIntegers.objects.create(plain=self.test_value)
        with self.assertRaises(IntegrityError):
            UniqueIntegers.objects.create(plain=self.test_value)

    def test_bigintegerfield(self):
        UniqueIntegers.objects.create(big=self.test_value)
        with self.assertRaises(IntegrityError):
            UniqueIntegers.objects.create(big=self.test_value)

    def test_positiveintegerfield(self):
        UniqueIntegers.objects.create(positive=self.test_value)
        with self.assertRaises(IntegrityError):
            UniqueIntegers.objects.create(positive=self.test_value)

    def test_positivebigintegerfield(self):
        UniqueIntegers.objects.create(positive_big=self.test_value)
        with self.assertRaises(IntegrityError):
            UniqueIntegers.objects.create(positive_big=self.test_value)



from bson.binary import Binary
from django.core.exceptions import ImproperlyConfigured
from django.db import connections

from .models import EncryptionKey, KeyTestModel, Patient
from .test_base import EncryptionTestCase


class SchemaTests(EncryptionTestCase):
    def setUp(self):
        self.reset_database_wrapper_encryption_cached_properties()

    def tearDown(self):
        self.reset_database_wrapper_encryption_cached_properties()

    def reset_database_wrapper_encryption_cached_properties(self):
        # Reset these cached properties before and after each test since some
        # tests mock values that form them.
        connection = connections["encrypted"]
        connection.__dict__.pop("kms_provider", None)
        connection.__dict__.pop("kms_credentials", None)

    def test_key_creation(self):
        """
        SchemaEditor._get_encrypted_fields() generates data encryption keys
        for the given model.
        """
        test_key_alt_name = f"{KeyTestModel._meta.db_table}.value"
        # Precondition: the key doesn't exist.
        with self.assertRaises(EncryptionKey.DoesNotExist):
            EncryptionKey.objects.get(key_alt_name=test_key_alt_name)
        # Generate the encryption key.
        with connections["encrypted"].schema_editor() as editor:
            encrypted_fields = editor._get_encrypted_fields(KeyTestModel)
        try:
            # The schema contains the new keyId.
            field_info = encrypted_fields["fields"][0]
            self.assertEqual(field_info["path"], "value")
            self.assertIsInstance(field_info["keyId"], Binary)
            # The key vault contains the new keyId.
            key = EncryptionKey.objects.get(key_alt_name=test_key_alt_name)
            self.assertEqual(key.id, field_info["keyId"])
            self.assertEqual(key.key_alt_name, [test_key_alt_name])
        finally:
            # Delete the test key.
            count, _ = EncryptionKey.objects.filter(key_alt_name=test_key_alt_name).delete()
            self.assertEqual(count, 1)

    def test_missing_auto_encryption_opts(self):
        connection = connections["default"]
        msg = (
            "Tried to create model encryption_.Patient in 'default' database. "
            "The model has encrypted fields but DATABASES['default']['OPTIONS'] "
            'is missing the "auto_encryption_opts" parameter. If the model '
            "should not be created in this database, adjust your database "
            "routers."
        )
        with (
            self.assertRaisesMessage(ImproperlyConfigured, msg),
            connection.schema_editor() as editor,
        ):
            editor.create_model(Patient)

    def test_missing_kms_credentials(self):
        connection = connections["encrypted"]
        auto_encryption_opts = connection.connection._options.auto_encryption_opts
        kms_providers = auto_encryption_opts._kms_providers
        # Mock a GCP KMS.
        auto_encryption_opts._kms_providers = {"gcp": {}}
        # Mock the lack of KMS_CREDENTIALS.
        kms_credentials = None
        if "KMS_CREDENTIALS" in connection.settings_dict:
            kms_credentials = connection.settings_dict.pop("KMS_CREDENTIALS")
        msg = "DATABASES['encrypted'] is missing 'KMS_CREDENTIALS' required for KMS 'gcp'."
        try:
            with (
                self.assertRaisesMessage(ImproperlyConfigured, msg),
                connection.schema_editor() as editor,
            ):
                editor.create_model(Patient)
        finally:
            # Restore the original values.
            auto_encryption_opts._kms_providers = kms_providers
            if kms_credentials is not None:
                connection.settings_dict["KMS_CREDENTIALS"] = kms_credentials

    def test_missing_kms_credentials_entry(self):
        connection = connections["encrypted"]
        auto_encryption_opts = connection.connection._options.auto_encryption_opts
        kms_providers = auto_encryption_opts._kms_providers
        # Mock a GCP KMS.
        auto_encryption_opts._kms_providers = {"gcp": {}}
        # Mock an empty KMS_CREDENTIALS.
        kms_credentials = None
        if "KMS_CREDENTIALS" in connection.settings_dict:
            kms_credentials = connection.settings_dict.pop("KMS_CREDENTIALS")
        connection.settings_dict["KMS_CREDENTIALS"] = {}
        msg = "DATABASES['encrypted']['KMS_CREDENTIALS'] is missing 'gcp' key."
        try:
            with (
                self.assertRaisesMessage(ImproperlyConfigured, msg),
                connection.schema_editor() as editor,
            ):
                editor.create_model(Patient)
        finally:
            # Restore the original values.
            auto_encryption_opts._kms_providers = kms_providers
            if kms_credentials is not None:
                connection.settings_dict["KMS_CREDENTIALS"] = kms_credentials
            else:
                connection.settings_dict.pop("KMS_CREDENTIALS")

    def test_multiple_kms_providers(self):
        connection = connections["encrypted"]
        auto_encryption_opts = connection.connection._options.auto_encryption_opts
        kms_providers = auto_encryption_opts._kms_providers
        # Mock multiple kms_providers by using a list of length > 1.
        auto_encryption_opts._kms_providers = [{}, {}]
        msg = (
            "Multiple KMS providers per database aren't supported. Please "
            "create a feature request with details about your use case."
        )
        try:
            with (
                self.assertRaisesMessage(ImproperlyConfigured, msg),
                connection.schema_editor() as editor,
            ):
                editor.create_model(Patient)
        finally:
            # Restore the original value.
            auto_encryption_opts._kms_providers = kms_providers



from pathlib import Path
from unittest import TestCase

from bson import encode, json_util

from .base import PerformanceTest
from .models import LargeFlatModel


class LargeFlatDocTest(PerformanceTest):
    """Parent class for large flat document tests."""

    dataset = "large_doc.json"

    def setUp(self):
        super().setUp()
        with open(  # noqa: PTH123
            Path(self.test_data_path) / Path("flat-models") / self.dataset
        ) as data:
            self.document = json_util.loads(data.read())
        self.setUpData()
        self.data_size = len(encode(self.document)) * self.num_docs

    def setUpData(self):
        LargeFlatModel.objects.bulk_create(
            LargeFlatModel(**self.document) for _ in range(self.num_docs)
        )

    def tearDown(self):
        super().tearDown()
        LargeFlatModel.objects.all().delete()


class TestLargeFlatDocCreation(LargeFlatDocTest, TestCase):
    """Creating a large flat document."""

    def setUpData(self):
        # Don't create data since this is the creation test.
        pass

    def do_task(self):
        for _ in range(self.num_docs):
            LargeFlatModel.objects.create(**self.document)

    def after(self):
        LargeFlatModel.objects.all().delete()


class TestLargeFlatDocUpdate(LargeFlatDocTest, TestCase):
    """Updating a field within a large flat document."""

    def setUp(self):
        super().setUp()
        self.models = list(LargeFlatModel.objects.all())
        self.data_size = len(encode({"field1": "updated_value0"})) * self.num_docs
        self.iteration = 0

    def do_task(self):
        for model in self.models:
            model.field1 = "updated_value" + str(self.iteration)
            model.save()
        self.iteration += 1


class TestLargeFlatDocFilterPkByIn(LargeFlatDocTest, TestCase):
    """Filtering large flat documents using __in for primary keys."""

    def setUp(self):
        super().setUp()
        self.ids = LargeFlatModel.objects.values_list("id", flat=True)[:100]

    def do_task(self):
        list(LargeFlatModel.objects.filter(id__in=self.ids))



from math import radians

from django.db import NotSupportedError


class Operator:
    def as_sql(self, connection, lookup, template_params, sql_params):
        # Return some dummy value to prevent str(queryset.query) from crashing.
        # The output of as_sql() is meaningless for this no-SQL backend.
        return self.name, []


class Contains(Operator):
    name = "contains"

    def as_mql(self, field, value, params=None):
        value_type = value["type"]
        if value_type != "Point":
            raise NotSupportedError(
                "MongoDB does not support contains on non-Point lookup geometries."
            )
        return {
            field: {
                "$geoIntersects": {
                    "$geometry": {
                        "type": value_type,
                        "coordinates": value["coordinates"],
                    }
                }
            }
        }


class Disjoint(Operator):
    name = "disjoint"

    def as_mql(self, field, value, params=None):
        return {
            field: {
                "$not": {
                    "$geoIntersects": {
                        "$geometry": {
                            "type": value["type"],
                            "coordinates": value["coordinates"],
                        }
                    }
                }
            }
        }


class DistanceLTE(Operator):
    name = "distance_lte"

    def get_geo_within(self, value, params):
        distance = params[0]
        # Get the distance in meters if it's a Distance object.
        distance = distance.m if hasattr(distance, "m") else distance
        return {
            "$geoWithin": {
                "$centerSphere": [
                    value["coordinates"],
                    distance / 6378100,  # radius of earth in meters
                ],
            }
        }

    def as_mql(self, field, value, params=None):
        return {field: self.get_geo_within(value, params)}


class DistanceGT(DistanceLTE):
    name = "distance_gt"

    def as_mql(self, field, value, params=None):
        return {field: {"$not": self.get_geo_within(value, params)}}


class DWithin(Operator):
    name = "dwithin"

    def as_mql(self, field, value, params=None):
        # The parameter is always in degrees. GISOperations.get_distance()
        # prohibits using Distance objects since MongoDB doesn't support any
        # projected systems.
        param = radians(params[0])
        return {field: {"$geoWithin": {"$centerSphere": [value["coordinates"], param]}}}


class Intersects(Operator):
    name = "intersects"

    def as_mql(self, field, value, params=None):
        return {
            field: {
                "$geoIntersects": {
                    "$geometry": {
                        "type": value["type"],
                        "coordinates": value["coordinates"],
                    }
                }
            }
        }


class Within(Operator):
    name = "within"

    def as_mql(self, field, value, params=None):
        return {
            field: {
                "$geoWithin": {
                    "$geometry": {
                        "type": value["type"],
                        "coordinates": value["coordinates"],
                    }
                }
            }
        }



import json
import logging
import os
import shlex
import subprocess
import sys
from datetime import datetime
from pathlib import Path

LOGGER = logging.getLogger("test")
logging.basicConfig(level=logging.INFO, format="%(levelname)-8s %(message)s")
OUTPUT_FILE = os.environ.get("OUTPUT_FILE", "results.json")


def create_report(start_time: datetime):
    """Format the output from the performance tests into a report.json file."""
    end_time = datetime.now()
    elapsed_secs = (end_time - start_time).total_seconds()
    with open(OUTPUT_FILE) as fid:  # noqa: PTH123
        results = json.load(fid)
    LOGGER.info("%s:\n%s", OUTPUT_FILE, json.dumps(results, indent=2))
    results = {
        "status": "PASS",
        "exit_code": 0,
        "test_file": "BenchmarkTests",
        "start": int(start_time.timestamp()),
        "end": int(end_time.timestamp()),
        "elapsed": elapsed_secs,
    }
    report = {"results": [results]}
    LOGGER.info("report.json\n%s", json.dumps(report, indent=2))
    with open("report.json", "w", newline="\n") as fid:  # noqa: PTH123
        json.dump(report, fid)


def run_command(cmd: str | list[str], **kwargs) -> None:
    """Run a shell command. Exit on failure."""
    if isinstance(cmd, list):
        cmd = " ".join(cmd)
    LOGGER.info("Running command '%s'...", cmd)
    kwargs.setdefault("check", True)
    try:
        subprocess.run(shlex.split(cmd), **kwargs)  # noqa: PLW1510, S603
    except subprocess.CalledProcessError as e:
        LOGGER.error(e.output)
        LOGGER.error(str(e))
        sys.exit(e.returncode)
    LOGGER.info("Running command '%s'... done.", cmd)


start_time = datetime.now()
ROOT = Path(__file__).absolute().parent.parent.parent
data_dir = ROOT / "specifications/source/benchmarking/odm-data"
if not data_dir.exists():
    run_command("git clone --depth 1 https://github.com/mongodb/specifications.git")
    run_command("tar xf flat_models.tgz", cwd=data_dir)
    run_command("tar xf nested_models.tgz", cwd=data_dir)

os.chdir("performance_tests")
start_time = datetime.now()
run_command(
    "python runtests.py",
    env=os.environ
    | {
        "DJANGO_MONGODB_PERFORMANCE_TEST_DATA_PATH": str(data_dir),
        "OUTPUT_FILE": OUTPUT_FILE,
    },
)
create_report(start_time)



from django.db import models


class Author(models.Model):
    name = models.CharField(max_length=50)

    def __str__(self):
        return self.name


class Blob(models.Model):
    name = models.CharField(max_length=10)
    data = models.JSONField(null=True)

    def __str__(self):
        return self.name



from django.core.exceptions import ImproperlyConfigured
from django.db import NotSupportedError, connections
from django.db.models import F

from .models import CharModel
from .test_base import EncryptionTestCase


class ModelTests(EncryptionTestCase):
    def test_create_in_non_encrypted_connection(self):
        connections["default"].close_pool()
        # Simulate the first connection to the database where cached properties
        # aren't initialized.
        del connections["default"].auto_encryption_opts
        msg = "Cannot save encrypted field 'value' in non-encrypted database 'default'."
        with self.assertRaisesMessage(ImproperlyConfigured, msg):
            CharModel.objects.using("default").create(value="value")

    def test_update(self):
        obj = CharModel.objects.create(value="hello")
        obj.value = "updated"
        obj.save()
        obj.refresh_from_db()
        self.assertEqual(obj.value, "updated")

    def test_update_with_expression(self):
        obj = CharModel.objects.create(value="hello")
        obj.value = F("value")
        msg = "Expressions in update queries are not allowed with Queryable Encryption."
        with self.assertRaisesMessage(NotSupportedError, msg):
            obj.save()
        obj.refresh_from_db()
        self.assertEqual(obj.value, "hello")



from bson import SON, ObjectId
from django.db import models
from django.db.models import Case, F, IntegerField, Value, When
from django.test import TestCase

from django_mongodb_backend.test import MongoTestCaseMixin

from .models import Author, Book, Library, Order, Tag


class MQLTests(MongoTestCaseMixin, TestCase):
    def test_all(self):
        with self.assertNumQueries(1) as ctx:
            list(Author.objects.all())
        self.assertAggregateQuery(ctx.captured_queries[0]["sql"], "queries__author", [])

    def test_join(self):
        with self.assertNumQueries(1) as ctx:
            list(Book.objects.filter(author__name="Bob"))
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__book",
            [
                {
                    "$lookup": {
                        "as": "queries__author",
                        "foreignField": "_id",
                        "from": "queries__author",
                        "localField": "author_id",
                        "pipeline": [{"$match": {"name": "Bob"}}],
                    }
                },
                {"$unwind": "$queries__author"},
                {"$match": {"queries__author.name": "Bob"}},
            ],
        )


class QuerySetQueryTests(TestCase):
    def test_all(self):
        self.assertEqual(str(Author.objects.all().query), "db.queries__author.aggregate([])")

    def test_filter(self):
        self.assertEqual(
            str(Author.objects.filter(name="Bob").query),
            "db.queries__author.aggregate([{'$match': {'name': 'Bob'}}])",
        )

    def test_values(self):
        self.assertEqual(
            str(Author.objects.values("name").query),
            "db.queries__author.aggregate([{'$project': {'name': 1}}])",
        )


class FKLookupConditionPushdownTests(MongoTestCaseMixin, TestCase):
    def test_filter_on_local_and_related_fields(self):
        with self.assertNumQueries(1) as ctx:
            list(Book.objects.filter(title="Don", author__name="John"))
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__book",
            [
                {
                    "$lookup": {
                        "from": "queries__author",
                        "pipeline": [{"$match": {"name": "John"}}],
                        "as": "queries__author",
                        "localField": "author_id",
                        "foreignField": "_id",
                    }
                },
                {"$unwind": "$queries__author"},
                {"$match": {"$and": [{"queries__author.name": "John"}, {"title": "Don"}]}},
            ],
        )

    def test_or_mixing_local_and_related_fields_is_not_pushable(self):
        with self.assertNumQueries(1) as ctx:
            list(Book.objects.filter(models.Q(title="Don") | models.Q(author__name="John")))
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__book",
            [
                {
                    "$lookup": {
                        "from": "queries__author",
                        "pipeline": [],
                        "as": "queries__author",
                        "localField": "author_id",
                        "foreignField": "_id",
                    }
                },
                {"$unwind": "$queries__author"},
                {"$match": {"$or": [{"title": "Don"}, {"queries__author.name": "John"}]}},
            ],
        )

    def test_filter_on_self_join_fields(self):
        with self.assertNumQueries(1) as ctx:
            list(
                Tag.objects.filter(
                    parent__name="parent", parent__group_id=ObjectId("6891ff7822e475eddc20f159")
                )
            )
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__tag",
            [
                {
                    "$lookup": {
                        "from": "queries__tag",
                        "pipeline": [
                            {
                                "$match": {
                                    "$and": [
                                        {"group_id": ObjectId("6891ff7822e475eddc20f159")},
                                        {"name": "parent"},
                                    ]
                                }
                            }
                        ],
                        "as": "T2",
                        "localField": "parent_id",
                        "foreignField": "_id",
                    }
                },
                {"$unwind": "$T2"},
                {
                    "$match": {
                        "$and": [
                            {"T2.group_id": ObjectId("6891ff7822e475eddc20f159")},
                            {"T2.name": "parent"},
                        ]
                    }
                },
            ],
        )

    def test_filter_on_reverse_foreignkey_relation(self):
        with self.assertNumQueries(1) as ctx:
            list(Order.objects.filter(items__status=ObjectId("6891ff7822e475eddc20f159")))
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__order",
            [
                {
                    "$lookup": {
                        "from": "queries__orderitem",
                        "localField": "_id",
                        "foreignField": "order_id",
                        "as": "queries__orderitem",
                        "pipeline": [{"$match": {"status": ObjectId("6891ff7822e475eddc20f159")}}],
                    }
                },
                {"$unwind": "$queries__orderitem"},
                {"$match": {"queries__orderitem.status": ObjectId("6891ff7822e475eddc20f159")}},
                {"$addFields": {"_id": "$_id"}},
                {"$sort": SON([("_id", 1)])},
            ],
        )

    def test_filter_on_local_and_nested_join_fields(self):
        with self.assertNumQueries(1) as ctx:
            list(
                Order.objects.filter(
                    name="My Order",
                    items__order__name="My Order",
                    items__status=ObjectId("6891ff7822e475eddc20f159"),
                )
            )
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__order",
            [
                {
                    "$lookup": {
                        "from": "queries__orderitem",
                        "localField": "_id",
                        "foreignField": "order_id",
                        "pipeline": [{"$match": {"status": ObjectId("6891ff7822e475eddc20f159")}}],
                        "as": "queries__orderitem",
                    }
                },
                {"$unwind": "$queries__orderitem"},
                {
                    "$lookup": {
                        "from": "queries__order",
                        "pipeline": [{"$match": {"name": "My Order"}}],
                        "as": "T3",
                        "localField": "queries__orderitem.order_id",
                        "foreignField": "_id",
                    }
                },
                {"$unwind": "$T3"},
                {
                    "$match": {
                        "$and": [
                            {"T3.name": "My Order"},
                            {"queries__orderitem.status": ObjectId("6891ff7822e475eddc20f159")},
                            {"name": "My Order"},
                        ]
                    }
                },
                {"$addFields": {"_id": "$_id"}},
                {"$sort": SON([("_id", 1)])},
            ],
        )

    def test_negated_related_filter_is_pushable(self):
        with self.assertNumQueries(1) as ctx:
            list(Book.objects.filter(~models.Q(author__name="John")))
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__book",
            [
                {
                    "$lookup": {
                        "from": "queries__author",
                        "as": "queries__author",
                        "localField": "author_id",
                        "foreignField": "_id",
                        "pipeline": [
                            {
                                "$match": {
                                    "$nor": [{"name": "John"}],
                                }
                            }
                        ],
                    }
                },
                {"$unwind": "$queries__author"},
                {"$match": {"$nor": [{"queries__author.name": "John"}]}},
            ],
        )

    def test_or_on_local_fields_only(self):
        with self.assertNumQueries(1) as ctx:
            list(Order.objects.filter(models.Q(name="A") | models.Q(name="B")))
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__order",
            [
                {"$match": {"$or": [{"name": "A"}, {"name": "B"}]}},
                {"$addFields": {"_id": "$_id"}},
                {"$sort": SON([("_id", 1)])},
            ],
        )

    def test_or_with_mixed_pushable_and_non_pushable_fields(self):
        with self.assertNumQueries(1) as ctx:
            list(Book.objects.filter(models.Q(author__name="John") | models.Q(title="Don")))
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__book",
            [
                {
                    "$lookup": {
                        "from": "queries__author",
                        "pipeline": [],
                        "as": "queries__author",
                        "localField": "author_id",
                        "foreignField": "_id",
                    }
                },
                {"$unwind": "$queries__author"},
                {"$match": {"$or": [{"queries__author.name": "John"}, {"title": "Don"}]}},
            ],
        )

    def test_push_equality_between_parent_and_child_fields(self):
        with self.assertNumQueries(1) as ctx:
            list(Order.objects.filter(items__status=models.F("id")))
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__order",
            [
                {
                    "$lookup": {
                        "from": "queries__orderitem",
                        "let": {"parent__field__0": "$_id"},
                        "pipeline": [
                            {"$match": {"$expr": {"$eq": ["$status", "$$parent__field__0"]}}}
                        ],
                        "as": "queries__orderitem",
                        "localField": "_id",
                        "foreignField": "order_id",
                    }
                },
                {"$unwind": "$queries__orderitem"},
                {"$match": {"$expr": {"$eq": ["$queries__orderitem.status", "$_id"]}}},
                {"$addFields": {"_id": "$_id"}},
                {"$sort": SON([("_id", 1)])},
            ],
        )

    def test_double_negation_pushdown(self):
        a1 = Author.objects.create(name="Alice")
        a2 = Author.objects.create(name="Bob")
        b1 = Book.objects.create(title="Book1", author=a1, isbn="111")
        Book.objects.create(title="Book2", author=a2, isbn="222")
        b3 = Book.objects.create(title="Book3", author=a1, isbn="333")
        with self.assertNumQueries(1) as ctx:
            self.assertSequenceEqual(
                Book.objects.filter(~(~models.Q(author__name="Alice") | models.Q(title="Book4"))),
                [b1, b3],
            )
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__book",
            [
                {
                    "$lookup": {
                        "from": "queries__author",
                        "pipeline": [{"$match": {"name": "Alice"}}],
                        "as": "queries__author",
                        "localField": "author_id",
                        "foreignField": "_id",
                    }
                },
                {"$unwind": "$queries__author"},
                {
                    "$match": {
                        "$nor": [
                            {
                                "$or": [
                                    {"$nor": [{"queries__author.name": "Alice"}]},
                                    {"title": "Book4"},
                                ]
                            }
                        ]
                    }
                },
            ],
        )

    def test_partial_or_pushdown(self):
        a1 = Author.objects.create(name="Alice")
        a2 = Author.objects.create(name="Bob")
        a3 = Author.objects.create(name="Charlie")
        b1 = Book.objects.create(title="B1", author=a1, isbn="111")
        b2 = Book.objects.create(title="B2", author=a2, isbn="111")
        Book.objects.create(title="B3", author=a3, isbn="222")
        with self.assertNumQueries(1) as ctx:
            self.assertSequenceEqual(
                Book.objects.filter(
                    models.Q(author__name="Alice")
                    | (models.Q(author__name="Bob") & models.Q(isbn="111"))
                ),
                [b1, b2],
            )
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__book",
            [
                {
                    "$lookup": {
                        "from": "queries__author",
                        "pipeline": [{"$match": {"$or": [{"name": "Alice"}, {"name": "Bob"}]}}],
                        "as": "queries__author",
                        "localField": "author_id",
                        "foreignField": "_id",
                    }
                },
                {"$unwind": "$queries__author"},
                {
                    "$match": {
                        "$or": [
                            {"queries__author.name": "Alice"},
                            {"$and": [{"queries__author.name": "Bob"}, {"isbn": "111"}]},
                        ]
                    }
                },
            ],
        )

    def test_multiple_ors_with_partial_pushdown(self):
        a1 = Author.objects.create(name="Alice")
        a2 = Author.objects.create(name="Bob")
        a3 = Author.objects.create(name="Charlie")
        a4 = Author.objects.create(name="David")
        b1 = Book.objects.create(title="B1", author=a1, isbn="111")
        b2 = Book.objects.create(title="B2", author=a1, isbn="222")
        b3 = Book.objects.create(title="B3", author=a2, isbn="333")
        b4 = Book.objects.create(title="B4", author=a3, isbn="333")
        Book.objects.create(title="B5", author=a4, isbn="444")
        with self.assertNumQueries(1) as ctx:
            self.assertSequenceEqual(
                Book.objects.filter(
                    models.Q(author__name="Alice") & (models.Q(isbn="111") | models.Q(isbn="222"))
                    | (models.Q(author__name="Bob") | models.Q(author__name="Charlie"))
                    & models.Q(isbn="333")
                ),
                [b1, b2, b3, b4],
            )
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__book",
            [
                {
                    "$lookup": {
                        "from": "queries__author",
                        "pipeline": [
                            {
                                "$match": {
                                    "$or": [
                                        {"name": "Alice"},
                                        {"$or": [{"name": "Bob"}, {"name": "Charlie"}]},
                                    ]
                                }
                            }
                        ],
                        "as": "queries__author",
                        "localField": "author_id",
                        "foreignField": "_id",
                    }
                },
                {"$unwind": "$queries__author"},
                {
                    "$match": {
                        "$or": [
                            {
                                "$and": [
                                    {"queries__author.name": "Alice"},
                                    {"$or": [{"isbn": "111"}, {"isbn": "222"}]},
                                ]
                            },
                            {
                                "$and": [
                                    {
                                        "$or": [
                                            {"queries__author.name": "Bob"},
                                            {"queries__author.name": "Charlie"},
                                        ]
                                    },
                                    {"isbn": "333"},
                                ]
                            },
                        ]
                    }
                },
            ],
        )

    def test_self_join_tag_three_levels_none_pushable(self):
        t1 = Tag.objects.create(name="T1")
        t2 = Tag.objects.create(name="T2", parent=t1)
        t3 = Tag.objects.create(name="T3", parent=t2)
        Tag.objects.create(name="T4", parent=t3)
        Tag.objects.create(name="T5", parent=t1)
        t6 = Tag.objects.create(name="T6", parent=t2)
        with self.assertNumQueries(1) as ctx:
            self.assertSequenceEqual(
                Tag.objects.filter(
                    models.Q(name="T1")
                    | models.Q(parent__name="T2")
                    | models.Q(parent__parent__name="T3")
                ),
                [t1, t3, t6],
            )
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__tag",
            # Django translates this kind of query into left outer join.
            [
                {
                    "$lookup": {
                        "from": "queries__tag",
                        "pipeline": [],
                        "as": "T2",
                        "localField": "parent_id",
                        "foreignField": "_id",
                    }
                },
                {
                    "$set": {
                        "T2": {
                            "$cond": {
                                "if": {
                                    "$or": [
                                        {"$eq": [{"$type": "$T2"}, "missing"]},
                                        {"$eq": [{"$size": "$T2"}, 0]},
                                    ]
                                },
                                "then": [{}],
                                "else": "$T2",
                            }
                        }
                    }
                },
                {"$unwind": "$T2"},
                {
                    "$lookup": {
                        "from": "queries__tag",
                        "pipeline": [],
                        "as": "T3",
                        "localField": "T2.parent_id",
                        "foreignField": "_id",
                    }
                },
                {
                    "$set": {
                        "T3": {
                            "$cond": {
                                "if": {
                                    "$or": [
                                        {"$eq": [{"$type": "$T3"}, "missing"]},
                                        {"$eq": [{"$size": "$T3"}, 0]},
                                    ]
                                },
                                "then": [{}],
                                "else": "$T3",
                            }
                        }
                    }
                },
                {"$unwind": "$T3"},
                {"$match": {"$or": [{"name": "T1"}, {"T2.name": "T2"}, {"T3.name": "T3"}]}},
            ],
        )

    def test_self_join_tag_three_levels_pushable(self):
        t1 = Tag.objects.create(name="T1")
        t2 = Tag.objects.create(name="T2", parent=t1)
        t3 = Tag.objects.create(name="T3", parent=t2)
        Tag.objects.create(name="T4", parent=t3)
        Tag.objects.create(name="T5", parent=t1)
        Tag.objects.create(name="T6", parent=t2)
        with self.assertNumQueries(1) as ctx:
            self.assertSequenceEqual(
                Tag.objects.filter(name="T1", parent__name="T2", parent__parent__name="T3"),
                [],
            )
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__tag",
            [
                {
                    "$lookup": {
                        "from": "queries__tag",
                        "pipeline": [{"$match": {"name": "T2"}}],
                        "as": "T2",
                        "localField": "parent_id",
                        "foreignField": "_id",
                    }
                },
                {"$unwind": "$T2"},
                {
                    "$lookup": {
                        "from": "queries__tag",
                        "pipeline": [{"$match": {"name": "T3"}}],
                        "as": "T3",
                        "localField": "T2.parent_id",
                        "foreignField": "_id",
                    }
                },
                {"$unwind": "$T3"},
                {"$match": {"$and": [{"name": "T1"}, {"T2.name": "T2"}, {"T3.name": "T3"}]}},
            ],
        )

    def test_partial_and_pushdown(self):
        a1 = Author.objects.create(name="Alice")
        a2 = Author.objects.create(name="Bob")
        b1 = Book.objects.create(title="B1", author=a1, isbn="111")
        Book.objects.create(title="B2", author=a2, isbn="222")
        with self.assertNumQueries(1) as ctx:
            self.assertSequenceEqual(
                Book.objects.filter(models.Q(author__name="Alice") & models.Q(title__contains="B")),
                [b1],
            )
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__book",
            [
                {
                    "$lookup": {
                        "from": "queries__author",
                        "pipeline": [{"$match": {"name": "Alice"}}],
                        "as": "queries__author",
                        "localField": "author_id",
                        "foreignField": "_id",
                    }
                },
                {"$unwind": "$queries__author"},
                {
                    "$match": {
                        "$and": [
                            {"queries__author.name": "Alice"},
                            {"title": {"$regex": "B", "$options": ""}},
                        ]
                    }
                },
            ],
        )

    def test_not_or_demorgan_pushdown(self):
        a1 = Author.objects.create(name="Alice")
        a2 = Author.objects.create(name="Bob")
        b1 = Book.objects.create(title="B1", author=a1, isbn="111")
        Book.objects.create(title="B2", author=a2, isbn="222")
        with self.assertNumQueries(1) as ctx:
            self.assertSequenceEqual(
                Book.objects.filter(~(models.Q(author__name="Bob") | models.Q(isbn="222"))),
                [b1],
            )
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__book",
            [
                {
                    "$lookup": {
                        "from": "queries__author",
                        "pipeline": [{"$match": {"$nor": [{"name": "Bob"}]}}],
                        "as": "queries__author",
                        "localField": "author_id",
                        "foreignField": "_id",
                    }
                },
                {"$unwind": "$queries__author"},
                {"$match": {"$nor": [{"$or": [{"queries__author.name": "Bob"}, {"isbn": "222"}]}]}},
            ],
        )

    def test_or_mixed_local_remote_pushdown(self):
        a1 = Author.objects.create(name="Alice")
        a2 = Author.objects.create(name="Bob")
        b1 = Book.objects.create(title="B1", author=a1, isbn="111")
        b2 = Book.objects.create(title="B2", author=a2, isbn="222")
        with self.assertNumQueries(1) as ctx:
            self.assertSequenceEqual(
                Book.objects.filter(models.Q(title="B1") | models.Q(author__name="Bob")), [b1, b2]
            )
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__book",
            [
                {
                    "$lookup": {
                        "from": "queries__author",
                        "pipeline": [],
                        "as": "queries__author",
                        "localField": "author_id",
                        "foreignField": "_id",
                    }
                },
                {"$unwind": "$queries__author"},
                {"$match": {"$or": [{"title": "B1"}, {"queries__author.name": "Bob"}]}},
            ],
        )

    def test_conditional_expression_not_pushed(self):
        a1 = Author.objects.create(name="Vicente")
        a2 = Author.objects.create(name="Carlos")
        a3 = Author.objects.create(name="Maria")
        Book.objects.create(title="B1", author=a1, isbn="111")
        b2 = Book.objects.create(title="B2", author=a2, isbn="222")
        Book.objects.create(title="B3", author=a3, isbn="333")
        qs = Book.objects.annotate(
            score=Case(
                When(author__name="Vicente", then=Value(2)),
                When(author__name="Carlos", then=Value(4)),
                default=Value(1),
                output_field=IntegerField(),
            )
            + Value(1)
        ).filter(score__gt=3)
        with self.assertNumQueries(1) as ctx:
            self.assertSequenceEqual(qs, [b2])
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__book",
            [
                {
                    "$lookup": {
                        "from": "queries__author",
                        "pipeline": [],
                        "as": "queries__author",
                        "localField": "author_id",
                        "foreignField": "_id",
                    }
                },
                {"$unwind": "$queries__author"},
                {
                    "$match": {
                        "$expr": {
                            "$gt": [
                                {
                                    "$add": [
                                        {
                                            "$switch": {
                                                "branches": [
                                                    {
                                                        "case": {
                                                            "$eq": [
                                                                "$queries__author.name",
                                                                "Vicente",
                                                            ]
                                                        },
                                                        "then": {"$literal": 2},
                                                    },
                                                    {
                                                        "case": {
                                                            "$eq": [
                                                                "$queries__author.name",
                                                                "Carlos",
                                                            ]
                                                        },
                                                        "then": {"$literal": 4},
                                                    },
                                                ],
                                                "default": {"$literal": 1},
                                            }
                                        },
                                        {"$literal": 1},
                                    ]
                                },
                                3,
                            ]
                        }
                    }
                },
                {
                    "$project": {
                        "score": {
                            "$add": [
                                {
                                    "$switch": {
                                        "branches": [
                                            {
                                                "case": {
                                                    "$eq": ["$queries__author.name", "Vicente"]
                                                },
                                                "then": {"$literal": 2},
                                            },
                                            {
                                                "case": {
                                                    "$eq": ["$queries__author.name", "Carlos"]
                                                },
                                                "then": {"$literal": 4},
                                            },
                                        ],
                                        "default": {"$literal": 1},
                                    }
                                },
                                {"$literal": 1},
                            ]
                        },
                        "_id": 1,
                        "title": 1,
                        "author_id": 1,
                        "isbn": 1,
                    }
                },
            ],
        )

    def test_simple_annotation_pushdown(self):
        a1 = Author.objects.create(name="Alice")
        a2 = Author.objects.create(name="Bob")
        b1 = Book.objects.create(title="B1", author=a1, isbn="111")
        Book.objects.create(title="B2", author=a2, isbn="222")
        b3 = Book.objects.create(title="B3", author=a1, isbn="333")
        with self.assertNumQueries(1) as ctx:
            self.assertSequenceEqual(
                Book.objects.annotate(name_length=F("author__name")).filter(name_length="Alice"),
                [b1, b3],
            )
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__book",
            [
                {
                    "$lookup": {
                        "from": "queries__author",
                        "pipeline": [{"$match": {"name": "Alice"}}],
                        "as": "queries__author",
                        "localField": "author_id",
                        "foreignField": "_id",
                    }
                },
                {"$unwind": "$queries__author"},
                {"$match": {"queries__author.name": "Alice"}},
                {
                    "$project": {
                        "queries__author": {"name_length": "$queries__author.name"},
                        "_id": 1,
                        "title": 1,
                        "author_id": 1,
                        "isbn": 1,
                    }
                },
            ],
        )


class M2MLookupConditionPushdownTests(MongoTestCaseMixin, TestCase):
    def test_simple_related_filter_is_pushed(self):
        with self.assertNumQueries(1) as ctx:
            list(Library.objects.filter(readers__name="Alice"))
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__library",
            [
                {
                    "$lookup": {
                        "from": "queries__library_readers",
                        "pipeline": [],
                        "as": "queries__library_readers",
                        "localField": "_id",
                        "foreignField": "library_id",
                    }
                },
                {"$unwind": "$queries__library_readers"},
                {
                    "$lookup": {
                        "from": "queries__reader",
                        "pipeline": [{"$match": {"name": "Alice"}}],
                        "as": "queries__reader",
                        "localField": "queries__library_readers.reader_id",
                        "foreignField": "_id",
                    }
                },
                {"$unwind": "$queries__reader"},
                {"$match": {"queries__reader.name": "Alice"}},
            ],
        )

    def test_subquery_join_is_pushed(self):
        with self.assertNumQueries(1) as ctx:
            list(Library.objects.filter(~models.Q(readers__name="Alice")))
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__library",
            [
                {
                    "$lookup": {
                        "as": "__subquery0",
                        "from": "queries__library_readers",
                        "let": {"parent__field__0": "$_id"},
                        "pipeline": [
                            {
                                "$lookup": {
                                    "as": "U2",
                                    "foreignField": "_id",
                                    "from": "queries__reader",
                                    "localField": "reader_id",
                                    "pipeline": [{"$match": {"name": "Alice"}}],
                                }
                            },
                            {"$unwind": "$U2"},
                            {
                                "$match": {
                                    "$and": [
                                        {"U2.name": "Alice"},
                                        {"$expr": {"$eq": ["$library_id", "$$parent__field__0"]}},
                                    ]
                                }
                            },
                            {"$project": {"a": {"$literal": 1}}},
                            {"$limit": 1},
                        ],
                    }
                },
                {
                    "$set": {
                        "__subquery0": {
                            "$cond": {
                                "if": {
                                    "$or": [
                                        {"$eq": [{"$type": "$__subquery0"}, "missing"]},
                                        {"$eq": [{"$size": "$__subquery0"}, 0]},
                                    ]
                                },
                                "then": {},
                                "else": {"$arrayElemAt": ["$__subquery0", 0]},
                            }
                        }
                    }
                },
                {
                    "$match": {
                        "$nor": [
                            {
                                "$expr": {
                                    "$eq": [
                                        {
                                            "$not": {
                                                "$or": [
                                                    {
                                                        "$eq": [
                                                            {"$type": "$__subquery0.a"},
                                                            "missing",
                                                        ]
                                                    },
                                                    {"$eq": ["$__subquery0.a", None]},
                                                ]
                                            }
                                        },
                                        True,
                                    ]
                                }
                            }
                        ]
                    }
                },
            ],
        )

    def test_filter_on_local_and_related_fields(self):
        with self.assertNumQueries(1) as ctx:
            list(Library.objects.filter(name="Central", readers__name="Alice"))
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__library",
            [
                {
                    "$lookup": {
                        "from": "queries__library_readers",
                        "localField": "_id",
                        "foreignField": "library_id",
                        "pipeline": [],
                        "as": "queries__library_readers",
                    }
                },
                {"$unwind": "$queries__library_readers"},
                {
                    "$lookup": {
                        "from": "queries__reader",
                        "pipeline": [{"$match": {"name": "Alice"}}],
                        "as": "queries__reader",
                        "localField": "queries__library_readers.reader_id",
                        "foreignField": "_id",
                    }
                },
                {"$unwind": "$queries__reader"},
                {"$match": {"$and": [{"name": "Central"}, {"queries__reader.name": "Alice"}]}},
            ],
        )

    def test_annotate_foreign_field(self):
        with self.assertNumQueries(1) as ctx:
            list(
                Library.objects.annotate(foreing_field=models.F("readers__name")).filter(
                    name="Ateneo"
                )
            )
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__library",
            [
                {
                    "$lookup": {
                        "from": "queries__library_readers",
                        "localField": "_id",
                        "foreignField": "library_id",
                        "pipeline": [],
                        "as": "queries__library_readers",
                    }
                },
                {
                    "$set": {
                        "queries__library_readers": {
                            "$cond": {
                                "if": {
                                    "$or": [
                                        {
                                            "$eq": [
                                                {"$type": "$queries__library_readers"},
                                                "missing",
                                            ]
                                        },
                                        {"$eq": [{"$size": "$queries__library_readers"}, 0]},
                                    ]
                                },
                                "then": [{}],
                                "else": "$queries__library_readers",
                            }
                        }
                    }
                },
                {"$unwind": "$queries__library_readers"},
                {
                    "$lookup": {
                        "from": "queries__reader",
                        "pipeline": [],
                        "as": "queries__reader",
                        "localField": "queries__library_readers.reader_id",
                        "foreignField": "_id",
                    }
                },
                {
                    "$set": {
                        "queries__reader": {
                            "$cond": {
                                "if": {
                                    "$or": [
                                        {"$eq": [{"$type": "$queries__reader"}, "missing"]},
                                        {"$eq": [{"$size": "$queries__reader"}, 0]},
                                    ]
                                },
                                "then": [{}],
                                "else": "$queries__reader",
                            }
                        }
                    }
                },
                {"$unwind": "$queries__reader"},
                {"$match": {"name": "Ateneo"}},
                {
                    "$project": {
                        "queries__reader": {"foreing_field": "$queries__reader.name"},
                        "_id": 1,
                        "name": 1,
                    }
                },
            ],
        )

    def test_or_with_mixed_pushable_and_non_pushable_fields(self):
        with self.assertNumQueries(1) as ctx:
            list(Library.objects.filter(models.Q(readers__name="Alice") | models.Q(name="Central")))
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__library",
            [
                {
                    "$lookup": {
                        "from": "queries__library_readers",
                        "localField": "_id",
                        "foreignField": "library_id",
                        "pipeline": [],
                        "as": "queries__library_readers",
                    }
                },
                {
                    "$set": {
                        "queries__library_readers": {
                            "$cond": {
                                "if": {
                                    "$or": [
                                        {
                                            "$eq": [
                                                {"$type": "$queries__library_readers"},
                                                "missing",
                                            ]
                                        },
                                        {"$eq": [{"$size": "$queries__library_readers"}, 0]},
                                    ]
                                },
                                "then": [{}],
                                "else": "$queries__library_readers",
                            }
                        }
                    }
                },
                {"$unwind": "$queries__library_readers"},
                {
                    "$lookup": {
                        "from": "queries__reader",
                        "pipeline": [],
                        "as": "queries__reader",
                        "localField": "queries__library_readers.reader_id",
                        "foreignField": "_id",
                    }
                },
                {
                    "$set": {
                        "queries__reader": {
                            "$cond": {
                                "if": {
                                    "$or": [
                                        {"$eq": [{"$type": "$queries__reader"}, "missing"]},
                                        {"$eq": [{"$size": "$queries__reader"}, 0]},
                                    ]
                                },
                                "then": [{}],
                                "else": "$queries__reader",
                            }
                        }
                    }
                },
                {"$unwind": "$queries__reader"},
                {"$match": {"$or": [{"queries__reader.name": "Alice"}, {"name": "Central"}]}},
            ],
        )



"""Tests for the MongoDB ODM Performance Benchmark Spec.

For instructions, see docs/internals/contributing/performance-tests.rst.
"""

import json
import os
import time
import warnings
from pathlib import Path

if os.environ.get("FASTBENCH"):
    # Each test will run for at least NUM_ITERATIONS and at least
    # MIN_ITERATION_TIME seconds.
    NUM_ITERATIONS = 1
    MIN_ITERATION_TIME = 5
    # A test times out (exiting early with a warning) if it runs for more than
    # MAX_ITERATION_TIME seconds but doesn't reach NUM_ITERATIONS.
    MAX_ITERATION_TIME = 10
    # The number of model instances to use for each test.
    NUM_DOCS = 1000
else:
    NUM_ITERATIONS = 2
    MIN_ITERATION_TIME = 30
    MAX_ITERATION_TIME = 300
    NUM_DOCS = 10000


class Timer:
    def __enter__(self):
        self.start = time.monotonic()
        return self

    def __exit__(self, *args):
        self.end = time.monotonic()
        self.interval = self.end - self.start


# The tearDown() method of each PerformanceTest appends metrics to this list.
# The test runner uses it at the end of the test suite to write the output
# file.
result_data = []


def write_output_file():
    if not result_data:
        return
    output = json.dumps(result_data, indent=4)
    if OUTPUT_FILE := os.environ.get("OUTPUT_FILE"):
        with open(OUTPUT_FILE, "w") as opf:  # noqa: PTH123
            opf.write(output)
    else:
        print(output)  # noqa: T201


# Copied from the driver benchmarking suite.
class PerformanceTest:
    dataset: str
    data_size: int
    test_data_path = os.environ.get(
        "DJANGO_MONGODB_PERFORMANCE_TEST_DATA_PATH",
        Path(os.path.realpath(__file__)).parent.parent / "odm-data",
    )
    num_docs = NUM_DOCS

    def setUp(self):
        self.setup_time = time.monotonic()

    def tearDown(self):
        duration = time.monotonic() - self.setup_time
        # Remove "Test" so that TestMyTestName is reported as "MyTestName".
        name = self.__class__.__name__[4:]
        median = self.percentile(50)
        if median is None:
            # Test failed.
            return
        megabytes_per_sec = self.data_size / median / 1000000
        print(  # noqa: T201
            f"Completed {self.__class__.__name__} {megabytes_per_sec:.3f} MB/s, "
            f"MEDIAN={self.percentile(50):.3f}s, "
            f"total time={duration:.3f}s, iterations={len(self.results)}"
        )
        result_data.append(
            {
                "info": {
                    "test_name": name,
                },
                "metrics": [
                    {
                        "name": "megabytes_per_sec",
                        "type": "MEDIAN",
                        "value": megabytes_per_sec,
                        "metadata": {
                            "improvement_direction": "up",
                            "measurement_unit": "megabytes_per_second",
                        },
                    },
                ],
            }
        )

    def before(self):
        pass

    def do_task(self):
        raise NotImplementedError

    def after(self):
        pass

    def percentile(self, percentile):
        if hasattr(self, "results"):
            sorted_results = sorted(self.results)
            percentile_index = int(len(sorted_results) * percentile / 100) - 1
            return sorted_results[percentile_index]
        # Test failed.
        return None

    def runTest(self):
        results = []
        start = time.monotonic()
        i = 0
        while True:
            i += 1
            self.before()
            with Timer() as timer:
                self.do_task()
            self.after()
            results.append(timer.interval)
            duration = time.monotonic() - start
            if duration > MIN_ITERATION_TIME and i >= NUM_ITERATIONS:
                break
            if duration > MAX_ITERATION_TIME:
                with warnings.catch_warnings():
                    warnings.simplefilter("default")
                    warnings.warn(
                        f"{self.__class__.__name__} timed out after {MAX_ITERATION_TIME}s, "
                        f"completed {i}/{NUM_ITERATIONS} iterations.",
                        stacklevel=2,
                    )
                break
        self.results = results



import contextlib

from django.core.exceptions import FieldDoesNotExist
from django.db.models.expressions import Col
from django.db.models.fields.related import lazy_related_operation
from django.db.models.lookups import Lookup, Transform

from . import PolymorphicEmbeddedModelField
from .array import ArrayField, ArrayLenTransform
from .embedded_model_array import (
    EmbeddedModelArrayFieldTransform,
    EmbeddedModelArrayFieldTransformFactory,
)
from .utils import serialize_model_reference


class PolymorphicEmbeddedModelArrayField(ArrayField):
    def __init__(self, embedded_models, **kwargs):
        if "size" in kwargs:
            raise ValueError("PolymorphicEmbeddedModelArrayField does not support size.")
        kwargs["editable"] = False
        super().__init__(PolymorphicEmbeddedModelField(embedded_models), **kwargs)
        self.embedded_models = embedded_models

    def contribute_to_class(self, cls, name, private_only=False, **kwargs):
        super().contribute_to_class(cls, name, private_only=private_only, **kwargs)

        if not cls._meta.abstract:
            # If embedded_models contains any strings, replace them with the actual
            # model classes.
            def _resolve_lookup(_, *resolved_models):
                self.embedded_models = resolved_models

            lazy_related_operation(_resolve_lookup, cls, *self.embedded_models)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if path == (
            "django_mongodb_backend.fields.polymorphic_embedded_model_array."
            "PolymorphicEmbeddedModelArrayField"
        ):
            path = "django_mongodb_backend.fields.PolymorphicEmbeddedModelArrayField"
        kwargs["embedded_models"] = tuple(
            serialize_model_reference(x) for x in self.embedded_models
        )
        del kwargs["base_field"]
        del kwargs["editable"]
        return name, path, args, kwargs

    def get_db_prep_value(self, value, connection, prepared=False):
        if isinstance(value, (list, tuple)):
            # Must call get_db_prep_save() rather than get_db_prep_value()
            # to transform model instances to dicts.
            return [self.base_field.get_db_prep_save(i, connection) for i in value]
        if value is not None:
            raise TypeError(
                f"Expected list of {self.embedded_models!r} instances, not {type(value)!r}."
            )
        return value

    def formfield(self, **kwargs):
        raise NotImplementedError("PolymorphicEmbeddedModelField does not support forms.")

    _get_model_from_label = PolymorphicEmbeddedModelField._get_model_from_label

    def get_transform(self, name):
        transform = super().get_transform(name)
        if transform:
            return transform
        for model in self.base_field.resolved_embedded_models:
            with contextlib.suppress(FieldDoesNotExist):
                field = model._meta.get_field(name)
                break
        else:
            raise FieldDoesNotExist(
                f"The models of field '{self.name}' have no field named '{name}'."
            )
        return PolymorphicArrayFieldTransformFactory(field)

    def _get_lookup(self, lookup_name):
        lookup = super()._get_lookup(lookup_name)
        if lookup is None or lookup is ArrayLenTransform:
            return lookup

        class EmbeddedModelArrayFieldLookups(Lookup):
            def as_mql(self, compiler, connection, as_expr=False):
                raise ValueError(
                    "Lookups aren't supported on PolymorphicEmbeddedModelArrayField. "
                    "Try querying one of its embedded fields instead."
                )

        return EmbeddedModelArrayFieldLookups


class PolymorphicArrayFieldTransform(EmbeddedModelArrayFieldTransform):
    field_class_name = "PolymorphicEmbeddedModelArrayField"

    def __init__(self, field, *args, **kwargs):
        # Skip EmbeddedModelArrayFieldTransform.__init__()
        Transform.__init__(self, *args, **kwargs)
        # Lookups iterate over the array of embedded models. A virtual column
        # of the queried field's type represents each element.
        column_target = field.clone()
        column_name = f"${self.VIRTUAL_COLUMN_ITERABLE}.{field.column}"
        column_target.name = f"{field.name}"
        column_target.db_column = column_name
        column_target.set_attributes_from_name(column_name)
        self._lhs = Col(None, column_target)
        self._sub_transform = None


class PolymorphicArrayFieldTransformFactory(EmbeddedModelArrayFieldTransformFactory):
    def __call__(self, *args, **kwargs):
        return PolymorphicArrayFieldTransform(self.field, *args, **kwargs)



from django.contrib.gis.geos import LineString, Point, Polygon
from django.contrib.gis.measure import Distance
from django.db import NotSupportedError
from django.db.models import Case, CharField, Value, When
from django.test import TestCase, skipUnlessDBFeature

from .models import City, Zipcode


@skipUnlessDBFeature("gis_enabled")
class LookupTests(TestCase):
    fixtures = ["initial"]

    def test_contains(self):
        qs = City.objects.filter(point__contains=Point(-95.363151, 29.763374)).values_list(
            "name", flat=True
        )
        self.assertCountEqual(qs, ["Houston"])

    def test_contains_errors_on_non_point(self):
        message = "MongoDB does not support contains on non-Point lookup geometries."
        with self.assertRaisesMessage(NotSupportedError, message):
            City.objects.filter(point__contains=LineString((0, 0), (1, 1))).first()

    def test_disjoint(self):
        qs = City.objects.filter(point__disjoint=Point(100, 50)).values_list("name", flat=True)
        self.assertCountEqual(
            qs,
            [
                "Houston",
                "Dallas",
                "Oklahoma City",
                "Wellington",
                "Pueblo",
                "Lawrence",
                "Chicago",
                "Victoria",
            ],
        )

    def test_distance_gt(self):
        houston = City.objects.get(name="Houston")
        qs = City.objects.filter(point__distance_gt=(houston.point, 362825)).values_list(
            "name", flat=True
        )
        self.assertCountEqual(
            qs,
            [
                "Dallas",
                "Oklahoma City",
                "Wellington",
                "Pueblo",
                "Lawrence",
                "Chicago",
                "Victoria",
            ],
        )

    def test_distance_lte(self):
        houston = City.objects.get(name="Houston")
        qs = City.objects.filter(point__distance_lte=(houston.point, 362826)).values_list(
            "name", flat=True
        )
        self.assertCountEqual(qs, ["Houston", "Dallas"])  # Dallas is roughly ~363 km from Houston

    def test_distance_units(self):
        chicago = City.objects.get(name="Chicago")
        qs = City.objects.filter(point__distance_lte=(chicago.point, Distance(km=720))).values_list(
            "name", flat=True
        )
        self.assertCountEqual(qs, ["Lawrence", "Chicago"])
        qs = City.objects.filter(point__distance_lte=(chicago.point, Distance(mi=447))).values_list(
            "name", flat=True
        )
        self.assertCountEqual(qs, ["Lawrence", "Chicago"])

    def test_dwithin(self):
        houston = City.objects.get(name="Houston")
        qs = City.objects.filter(point__dwithin=(houston.point, 11.46)).values_list(
            "name", flat=True
        )
        self.assertCountEqual(qs, ["Houston", "Dallas", "Pueblo", "Oklahoma City", "Lawrence"])

    def test_dwithin_unsupported_units(self):
        message = "Only numeric values of degree units are allowed on dwithin queries."
        with self.assertRaisesMessage(ValueError, message):
            City.objects.filter(point__dwithin=(Point(40.7670, -73.9820), Distance(km=1))).first()

    def test_intersects(self):
        city = City.objects.create(point=Point(95, 30))
        qs = City.objects.filter(point__intersects=Point(95, 30).buffer(10))
        self.assertCountEqual(qs, [city])

    def test_within(self):
        zipcode = Zipcode.objects.get(code="77002")
        qs = City.objects.filter(point__within=zipcode.poly).values_list("name", flat=True)
        self.assertCountEqual(qs, ["Houston"])

    def test_unsupported(self):
        msg = "MongoDB does not support the 'same_as' lookup."
        with self.assertRaisesMessage(NotSupportedError, msg):
            City.objects.get(point__same_as=Point(95, 30))

    def test_lookup_expression(self):
        downtown_area = Polygon(
            (
                (-122.4194, 37.7749),
                (-122.4194, 37.8049),
                (-122.3894, 37.8049),
                (-122.3894, 37.7749),
                (-122.4194, 37.7749),
            )
        )
        msg = "MongoDB does not support expressions for spatial lookup values."
        with self.assertRaisesMessage(NotSupportedError, msg):
            City.objects.annotate(
                area_type=Case(
                    When(point__within=downtown_area, then=Value("Downtown")),
                    default=Value("Other"),
                    output_field=CharField(),
                )
            ).first()



from io import StringIO

from bson import json_util
from django.core.exceptions import ImproperlyConfigured
from django.core.management import call_command
from django.test import modify_settings

from .models import KeyTestModel
from .test_base import EncryptionTestCase


@modify_settings(INSTALLED_APPS={"prepend": "django_mongodb_backend"})
class ShowEncryptedFieldsTests(EncryptionTestCase):
    # Expected encrypted field maps for all models with encrypted fields.
    expected_maps = {
        "encryption__arraymodel": {"fields": [{"bsonType": "array", "path": "values"}]},
        "encryption__patient": {
            "fields": [
                {
                    "bsonType": "string",
                    "path": "patient_record.ssn",
                    "queries": {"queryType": "equality"},
                },
                {"bsonType": "object", "path": "patient_record.billing"},
            ]
        },
        "encryption__movie": {"fields": [{"bsonType": "array", "path": "cast"}]},
        # Equality-queryable fields
        "encryption__binarymodel": {
            "fields": [
                {"bsonType": "binData", "path": "value", "queries": {"queryType": "equality"}}
            ]
        },
        "encryption__booleanmodel": {
            "fields": [{"bsonType": "bool", "path": "value", "queries": {"queryType": "equality"}}]
        },
        "encryption__charmodel": {
            "fields": [
                {"bsonType": "string", "path": "value", "queries": {"queryType": "equality"}}
            ]
        },
        "encryption__emailmodel": {
            "fields": [
                {"bsonType": "string", "path": "value", "queries": {"queryType": "equality"}}
            ]
        },
        "encryption__genericipaddressmodel": {
            "fields": [
                {"bsonType": "string", "path": "value", "queries": {"queryType": "equality"}}
            ]
        },
        "encryption__objectidmodel": {
            "fields": [
                {"bsonType": "objectId", "path": "value", "queries": {"queryType": "equality"}}
            ]
        },
        "encryption__textmodel": {
            "fields": [
                {"bsonType": "string", "path": "value", "queries": {"queryType": "equality"}}
            ]
        },
        "encryption__urlmodel": {
            "fields": [
                {"bsonType": "string", "path": "value", "queries": {"queryType": "equality"}}
            ]
        },
        "encryption__uuidmodel": {
            "fields": [
                {"bsonType": "string", "path": "value", "queries": {"queryType": "equality"}}
            ]
        },
        # Range-queryable fields
        "encryption__bigintegermodel": {
            "fields": [{"bsonType": "long", "path": "value", "queries": {"queryType": "range"}}]
        },
        "encryption__datemodel": {
            "fields": [{"bsonType": "date", "path": "value", "queries": {"queryType": "range"}}]
        },
        "encryption__datetimemodel": {
            "fields": [{"bsonType": "date", "path": "value", "queries": {"queryType": "range"}}]
        },
        "encryption__decimalmodel": {
            "fields": [{"bsonType": "decimal", "path": "value", "queries": {"queryType": "range"}}]
        },
        "encryption__durationmodel": {
            "fields": [{"bsonType": "long", "path": "value", "queries": {"queryType": "range"}}]
        },
        "encryption__floatmodel": {
            "fields": [{"bsonType": "double", "path": "value", "queries": {"queryType": "range"}}]
        },
        "encryption__integermodel": {
            "fields": [{"bsonType": "long", "path": "value", "queries": {"queryType": "range"}}]
        },
        "encryption__positivebigintegermodel": {
            "fields": [{"bsonType": "long", "path": "value", "queries": {"queryType": "range"}}]
        },
        "encryption__positiveintegermodel": {
            "fields": [{"bsonType": "long", "path": "value", "queries": {"queryType": "range"}}]
        },
        "encryption__positivesmallintegermodel": {
            "fields": [{"bsonType": "int", "path": "value", "queries": {"queryType": "range"}}]
        },
        "encryption__smallintegermodel": {
            "fields": [{"bsonType": "int", "path": "value", "queries": {"queryType": "range"}}]
        },
        "encryption__timemodel": {
            "fields": [{"bsonType": "date", "path": "value", "queries": {"queryType": "range"}}]
        },
    }

    def test_ouput(self):
        """The expected output appears for each model."""
        out = StringIO()
        call_command("showencryptedfieldsmap", "--database", "encrypted", verbosity=0, stdout=out)
        parsed_output = json_util.loads(out.getvalue())
        # All models accounted for and no extra models.
        self.assertEqual(len(parsed_output), len(self.expected_maps))
        for model_key, expected in self.expected_maps.items():
            with self.subTest(model=model_key):
                self.assertIn(model_key, parsed_output)
                model_data = parsed_output[model_key]
                for field in model_data["fields"]:
                    del field["keyId"]  # Can't compare dynamic value
                self.assertEqual(expected, model_data)

    def test_missing_key(self):
        """
        ImproperlyConfigured is raised when an encryption key isn't found for a
        model.
        """
        msg = (
            "Encryption key 'encryption__keytestmodel.value' not found. Have "
            "you migrated the encryption_.KeyTestModel model?"
        )
        # showencryptedfieldsmap ignores unmanaged models, so fake that this
        # model is managed.
        KeyTestModel._meta.managed = True
        try:
            with self.assertRaisesMessage(ImproperlyConfigured, msg):
                call_command("showencryptedfieldsmap", "--database", "encrypted", verbosity=0)
        finally:
            # Restore original state.
            KeyTestModel._meta.managed = False



from django.db import NotSupportedError
from django.db.models import CharField, Expression, FloatField, TextField
from django.db.models.expressions import F, Value
from django.db.models.lookups import Lookup

from django_mongodb_backend.query_utils import process_lhs, process_rhs


def cast_as_field(path):
    return F(path) if isinstance(path, str) else path


class Operator:
    AND = "AND"
    OR = "OR"
    NOT = "NOT"

    def __init__(self, operator):
        self.operator = operator

    def __eq__(self, other):
        if isinstance(other, str):
            return self.operator == other
        return self.operator == other.operator

    def negate(self):
        if self.operator == self.AND:
            return Operator(self.OR)
        if self.operator == self.OR:
            return Operator(self.AND)
        return Operator(self.operator)

    def __hash__(self):
        return hash(self.operator)

    def __str__(self):
        return self.operator

    def __repr__(self):
        return self.operator


class SearchCombinable:
    def _combine(self, other, connector):
        if not isinstance(self, (CompoundExpression, CombinedSearchExpression)):
            lhs = CompoundExpression(must=[self])
        else:
            lhs = self
        if other and not isinstance(other, (CompoundExpression, CombinedSearchExpression)):
            rhs = CompoundExpression(must=[other])
        else:
            rhs = other
        return CombinedSearchExpression(lhs, connector, rhs)

    def __invert__(self):
        return self._combine(None, Operator(Operator.NOT))

    def __and__(self, other):
        return self._combine(other, Operator(Operator.AND))

    def __rand__(self, other):
        return self._combine(other, Operator(Operator.AND))

    def __or__(self, other):
        return self._combine(other, Operator(Operator.OR))

    def __ror__(self, other):
        return self._combine(other, Operator(Operator.OR))


class SearchExpression(SearchCombinable, Expression):
    """
    Base expression for $search stage.

    Subclasses produce the operator document placed under $search and expose
    the stage to queryset methods such as annotate(), filter(), or order_by().
    """

    output_field = FloatField()

    def __str__(self):
        cls = self.identity[0]
        kwargs = dict(self.identity[1:])
        arg_str = ", ".join(f"{k}={v!r}" for k, v in kwargs.items())
        return f"<{cls.__name__}({arg_str})>"

    def __repr__(self):
        return str(self)

    def as_sql(self, compiler, connection):
        return "", []

    def get_source_expressions(self):
        return [self.path]

    def set_source_expressions(self, exprs):
        (self.path,) = exprs

    def _get_indexed_fields(self, mappings):
        if isinstance(mappings, list):
            for definition in mappings:
                yield from self._get_indexed_fields(definition)
        else:
            for field, definition in mappings.get("fields", {}).items():
                yield field
                for path in self._get_indexed_fields(definition):
                    yield f"{field}.{path}"

    def _get_query_index(self, fields, compiler):
        fields = set(fields)
        for search_indexes in compiler.collection.list_search_indexes():
            mappings = search_indexes["latestDefinition"]["mappings"]
            indexed_fields = set(self._get_indexed_fields(mappings))
            if mappings["dynamic"] or fields.issubset(indexed_fields):
                return search_indexes["name"]
        return "default"

    def search_operator(self, compiler, connection):
        raise NotImplementedError

    def as_mql(self, compiler, connection, as_expr=False):
        index = self._get_query_index(self.get_search_fields(compiler, connection), compiler)
        return {"$search": {**self.search_operator(compiler, connection), "index": index}}


class SearchAutocomplete(SearchExpression):
    """
    Match input using the `autocomplete` operator.

    Enables autocomplete behavior by querying against a field indexed as
    `"type": "autocomplete"`.

    Example:
        SearchAutocomplete("title", "harry", fuzzy={"maxEdits": 1})

    Args:
        path: The document path to search (as string or expression).
        query: The input string to autocomplete.
        fuzzy: Optional dictionary of fuzzy matching parameters.
        token_order: Optional value for `"tokenOrder"`; controls sequential vs.
                     any-order token matching.
        score: Optional[SearchScore] expression to adjust score relevance
               (e.g., `{"boost": {"value": 5}}`).

    Reference: https://www.mongodb.com/docs/atlas/atlas-search/autocomplete/
    """

    def __init__(self, path, query, *, fuzzy=None, token_order=None, score=None):
        self.path = cast_as_field(path)
        self.query = query
        self.fuzzy = fuzzy
        self.token_order = token_order
        self.score = score
        super().__init__()

    def get_search_fields(self, compiler, connection):
        return {self.path.as_mql(compiler, connection)}

    def search_operator(self, compiler, connection):
        params = {
            "path": self.path.as_mql(compiler, connection),
            "query": self.query,
        }
        if self.score:
            params["score"] = self.score.as_mql(compiler, connection)
        if self.fuzzy is not None:
            params["fuzzy"] = self.fuzzy
        if self.token_order:
            params["tokenOrder"] = self.token_order
        return {"autocomplete": params}


class SearchEquals(SearchExpression):
    """
    Match documents with a field equal to the given value (`equals` operator).

    Example:
        SearchEquals("category", "fiction")

    Args:
        path: The document path to compare (as string or expression).
        value: The exact value to match against.
        score: Optional[SearchScore] expression to modify the relevance score.

    Reference: https://www.mongodb.com/docs/atlas/atlas-search/equals/
    """

    def __init__(self, path, value, *, score=None):
        self.path = cast_as_field(path)
        self.value = value
        self.score = score
        super().__init__()

    def get_search_fields(self, compiler, connection):
        return {self.path.as_mql(compiler, connection)}

    def search_operator(self, compiler, connection):
        params = {
            "path": self.path.as_mql(compiler, connection),
            "value": self.value,
        }
        if self.score:
            params["score"] = self.score.as_mql(compiler, connection)
        return {"equals": params}


class SearchExists(SearchExpression):
    """
    Match documents where a field exists.

    Use the `exists` operator to check whether a given path is present in the
    document. Useful for filtering documents that include (or exclude) optional
    fields.

    Example:
        SearchExists("metadata__author")

    Args:
        path: The document path to check (as string or expression).
        score: Optional[SearchScore] expression to modify the relevance score.

    Reference: https://www.mongodb.com/docs/atlas/atlas-search/exists/
    """

    def __init__(self, path, *, score=None):
        self.path = cast_as_field(path)
        self.score = score
        super().__init__()

    def get_search_fields(self, compiler, connection):
        return {self.path.as_mql(compiler, connection)}

    def search_operator(self, compiler, connection):
        params = {
            "path": self.path.as_mql(compiler, connection),
        }
        if self.score:
            params["score"] = self.score.as_mql(compiler, connection)
        return {"exists": params}


class SearchIn(SearchExpression):
    """
    Match documents where the field value is in a given list (`in` operator).

    Example:
        SearchIn("status", ["pending", "approved", "rejected"])

    Args:
        path: The document path to match against (as string or expression).
        value: A list of values to check for membership.
        score: Optional[SearchScore] expression to adjust the relevance score.

    Reference: https://www.mongodb.com/docs/atlas/atlas-search/in/
    """

    def __init__(self, path, value, *, score=None):
        self.path = cast_as_field(path)
        self.value = value
        self.score = score
        super().__init__()

    def get_search_fields(self, compiler, connection):
        return {self.path.as_mql(compiler, connection)}

    def search_operator(self, compiler, connection):
        params = {
            "path": self.path.as_mql(compiler, connection),
            "value": self.value,
        }
        if self.score:
            params["score"] = self.score.as_mql(compiler, connection)
        return {"in": params}


class SearchPhrase(SearchExpression):
    """
    Match a phrase in the specified field.

    Use the `phrase` operator to search for exact or near exact sequences of
    terms. It supports optional slop (word distance) and synonym sets.

    Example:
        SearchPhrase("description__text", "climate change", slop=2)

    Args:
        path: The document path to search (as string or expression).
        query: The phrase to match as a single string or list of terms.
        slop: Optional maximum word distance allowed between phrase terms.
        synonyms: Optional name of a synonym mapping defined in a search index.
        score: Optional[SearchScore] expression to modify the relevance score.

    Reference: https://www.mongodb.com/docs/atlas/atlas-search/phrase/
    """

    def __init__(self, path, query, *, slop=None, synonyms=None, score=None):
        self.path = cast_as_field(path)
        self.query = query
        self.slop = slop
        self.synonyms = synonyms
        self.score = score
        super().__init__()

    def get_search_fields(self, compiler, connection):
        return {self.path.as_mql(compiler, connection)}

    def search_operator(self, compiler, connection):
        params = {
            "path": self.path.as_mql(compiler, connection),
            "query": self.query,
        }
        if self.score:
            params["score"] = self.score.as_mql(compiler, connection)
        if self.slop:
            params["slop"] = self.slop
        if self.synonyms:
            params["synonyms"] = self.synonyms
        return {"phrase": params}


class SearchQueryString(SearchExpression):
    """
    Match using a Lucene-style query string.

    Use the `queryString` operator to parse and execute full-text queries
    written in a simplified Lucene syntax. It supports advanced constructs like
    boolean operators, wildcards, and field-specific terms.

    Example:
        SearchQueryString("content__text", "django AND (search OR query)")

    Args:
        path: The document path to query (as string or expression).
        query: The Lucene-style query string.
        score: Optional[SearchScore] expression to modify the relevance score.

    Reference: https://www.mongodb.com/docs/atlas/atlas-search/queryString/
    """

    def __init__(self, path, query, *, score=None):
        self.path = cast_as_field(path)
        self.query = query
        self.score = score
        super().__init__()

    def get_search_fields(self, compiler, connection):
        return {self.path.as_mql(compiler, connection)}

    def search_operator(self, compiler, connection):
        params = {
            "defaultPath": self.path.as_mql(compiler, connection),
            "query": self.query,
        }
        if self.score:
            params["score"] = self.score.as_mql(compiler, connection)
        return {"queryString": params}


class SearchRange(SearchExpression):
    """
    Filter documents within a range of values.

    Uses the `range` operator to match numeric, date, or other comparable
    fields based on upper and/or lower bounds.

    Example:
        SearchRange("published__year", gte=2000, lt=2020)

    Args:
        path: The document path to filter (as string or expression).
        lt: Optional exclusive upper bound (<).
        lte: Optional inclusive upper bound (≤).
        gt: Optional exclusive lower bound (>).
        gte: Optional inclusive lower bound (≥).
        score: Optional[SearchScore] expression to modify the relevance score.

    Reference: https://www.mongodb.com/docs/atlas/atlas-search/range/
    """

    def __init__(self, path, *, lt=None, lte=None, gt=None, gte=None, score=None):
        self.path = cast_as_field(path)
        self.lt = lt
        self.lte = lte
        self.gt = gt
        self.gte = gte
        self.score = score
        super().__init__()

    def get_search_fields(self, compiler, connection):
        return {self.path.as_mql(compiler, connection)}

    def search_operator(self, compiler, connection):
        params = {
            "path": self.path.as_mql(compiler, connection),
        }
        if self.score:
            params["score"] = self.score.as_mql(compiler, connection)
        if self.lt:
            params["lt"] = self.lt
        if self.lte:
            params["lte"] = self.lte
        if self.gt:
            params["gt"] = self.gt
        if self.gte:
            params["gte"] = self.gte
        return {"range": params}


class SearchRegex(SearchExpression):
    """
    Match strings using a regular expression (`regex` operator).

    Example:
        SearchRegex("username", r"^admin_")

    Args:
        path: The document path to match (as string or expression).
        query: The regular expression pattern to apply.
        allow_analyzed_field: Whether to allow matching against analyzed fields.
                              The server's default is False.
        score: Optional[SearchScore] expression to modify the relevance score.

    Reference: https://www.mongodb.com/docs/atlas/atlas-search/regex/
    """

    def __init__(self, path, query, *, allow_analyzed_field=None, score=None):
        self.path = cast_as_field(path)
        self.query = query
        self.allow_analyzed_field = allow_analyzed_field
        self.score = score
        super().__init__()

    def get_search_fields(self, compiler, connection):
        return {self.path.as_mql(compiler, connection)}

    def search_operator(self, compiler, connection):
        params = {
            "path": self.path.as_mql(compiler, connection),
            "query": self.query,
        }
        if self.score:
            params["score"] = self.score.as_mql(compiler, connection)
        if self.allow_analyzed_field is not None:
            params["allowAnalyzedField"] = self.allow_analyzed_field
        return {"regex": params}


class SearchText(SearchExpression):
    """
    Perform full-text search using the `text` operator.

    Matches terms in a specified field with options for fuzzy matching, match
    criteria, and synonyms.

    Example:
        SearchText(
            "description__content",
            "mongodb",
            fuzzy={"maxEdits": 1},
            match_criteria="all",
        )

    Args:
        path: The document path to search (as string or expression).
        query: The search term or phrase.
        fuzzy: Optional dictionary to configure fuzzy matching parameters.
        match_criteria: Optional criteria for term matching (e.g., "all" or "any").
        synonyms: Optional name of a synonym mapping defined in a search index.
        score: Optional[SearchScore] expression to adjust the relevance score.

    Reference: https://www.mongodb.com/docs/atlas/atlas-search/text/
    """

    def __init__(self, path, query, *, fuzzy=None, match_criteria=None, synonyms=None, score=None):
        self.path = cast_as_field(path)
        self.query = query
        self.fuzzy = fuzzy
        self.match_criteria = match_criteria
        self.synonyms = synonyms
        self.score = score
        super().__init__()

    def get_search_fields(self, compiler, connection):
        return {self.path.as_mql(compiler, connection)}

    def search_operator(self, compiler, connection):
        params = {
            "path": self.path.as_mql(compiler, connection),
            "query": self.query,
        }
        if self.score:
            params["score"] = self.score.as_mql(compiler, connection)
        if self.fuzzy is not None:
            params["fuzzy"] = self.fuzzy
        if self.match_criteria:
            params["matchCriteria"] = self.match_criteria
        if self.synonyms:
            params["synonyms"] = self.synonyms
        return {"text": params}


class SearchWildcard(SearchExpression):
    """
    Match strings using wildcard patterns.

    Use the `wildcard` operator to search for terms matching a pattern with
    `*` and `?` wildcards.

    Example:
        SearchWildcard("filename", "report_202?_final*")

    Args:
        path: The document path to search (as string or expression).
        query: The wildcard pattern to match.
        allow_analyzed_field: Whether to allow matching against analyzed fields.
                              The server's default is False.
        score: Optional[SearchScore] expression to modify the relevance score.

    Reference: https://www.mongodb.com/docs/atlas/atlas-search/wildcard/
    """

    def __init__(self, path, query, allow_analyzed_field=None, score=None):
        self.path = cast_as_field(path)
        self.query = query
        self.allow_analyzed_field = allow_analyzed_field
        self.score = score
        super().__init__()

    def get_search_fields(self, compiler, connection):
        return {self.path.as_mql(compiler, connection)}

    def search_operator(self, compiler, connection):
        params = {
            "path": self.path.as_mql(compiler, connection),
            "query": self.query,
        }
        if self.score:
            params["score"] = self.score.as_mql(compiler, connection)
        if self.allow_analyzed_field is not None:
            params["allowAnalyzedField"] = self.allow_analyzed_field
        return {"wildcard": params}


class SearchGeoShape(SearchExpression):
    """
    Filter documents by spatial relationship with a geometry.

    Uses the `geoShape` operator to match documents where a geo field relates
    to a specified geometry by a spatial relation.

    Example:
        SearchGeoShape(
            "location",
            "within",
            {"type": "Polygon", "coordinates": [...]},
        )

    Args:
        path: The document path to the geo field (as string or expression).
        relation: The spatial relation to test (e.g., "within", "intersects",
                  "disjoint").
        geometry: The GeoJSON geometry to compare against.
        score: Optional[SearchScore] expression to modify the relevance score.

    Reference: https://www.mongodb.com/docs/atlas/atlas-search/geoShape/
    """

    def __init__(self, path, relation, geometry, *, score=None):
        self.path = cast_as_field(path)
        self.relation = relation
        self.geometry = geometry
        self.score = score
        super().__init__()

    def get_search_fields(self, compiler, connection):
        return {self.path.as_mql(compiler, connection)}

    def search_operator(self, compiler, connection):
        params = {
            "path": self.path.as_mql(compiler, connection),
            "relation": self.relation,
            "geometry": self.geometry,
        }
        if self.score:
            params["score"] = self.score.as_mql(compiler, connection)
        return {"geoShape": params}


class SearchGeoWithin(SearchExpression):
    """
    Filter documents with geo fields contained within a specified shape.

    Use the `geoWithin` operator to match documents where the geo field lies
    entirely within the given geometry.

    Example:
        SearchGeoWithin(
            "location",
            "Polygon",
            {"type": "Polygon", "coordinates": [...]},
        )

    Args:
        path: The document path to the geo field (as string or expression).
        kind: The GeoJSON geometry type (e.g., "Polygon", "MultiPolygon").
        geometry: The GeoJSON geometry defining the boundary.
        score: Optional[SearchScore] expression to adjust the relevance score.

    Reference: https://www.mongodb.com/docs/atlas/atlas-search/geoWithin/
    """

    def __init__(self, path, kind, geometry, *, score=None):
        self.path = cast_as_field(path)
        self.kind = kind
        self.geometry = geometry
        self.score = score
        super().__init__()

    def get_search_fields(self, compiler, connection):
        return {self.path.as_mql(compiler, connection)}

    def search_operator(self, compiler, connection):
        params = {
            "path": self.path.as_mql(compiler, connection),
            self.kind: self.geometry,
        }
        if self.score:
            params["score"] = self.score.as_mql(compiler, connection)
        return {"geoWithin": params}


class SearchMoreLikeThis(SearchExpression):
    """
    Find documents similar to given examples.

    Use the `moreLikeThis` operator to search for documents that resemble the
    specified sample documents.

    Example:
        SearchMoreLikeThis([{"_id": ObjectId("...")}, {"title": "Example"}])

    Args:
        documents: A list of example documents or expressions to find similar
                   documents.
        score: Optional[SearchScore] expression to modify the relevance score.

    Reference: https://www.mongodb.com/docs/atlas/atlas-search/morelikethis/
    """

    def __init__(self, documents, *, score=None):
        self.documents = documents
        self.score = score
        super().__init__()

    def get_source_expressions(self):
        return []

    def set_source_expressions(self, exprs):
        pass

    def search_operator(self, compiler, connection):
        params = {
            "like": self.documents.as_mql(compiler, connection),
        }
        if self.score:
            params["score"] = self.score.as_mql(compiler, connection)
        return {"moreLikeThis": params}

    def get_search_fields(self, compiler, connection):
        needed_fields = set()
        for doc in self.documents.value:
            needed_fields.update(set(doc.keys()))
        return needed_fields


class CompoundExpression(SearchExpression):
    """
    Combine multiple search clauses using boolean logic.

    Use the `compound` operator to allow fine-grained control by combining
    multiple sub-expressions with `must`, `must_not`, `should`, and `filter`
    clauses.

    Example:
        CompoundExpression(
            must=[expr1, expr2],
            must_not=[expr3],
            should=[expr4],
            minimum_should_match=1,
        )

    Args:
        must: List of expressions that **must** match.
        must_not: List of expressions that **must not** match.
        should: List of expressions that **should** match (optional relevance
                boost).
        filter: List of expressions to filter results without affecting
                relevance.
        score: Optional[SearchScore] expression to adjust scoring.
        minimum_should_match: Minimum number of `should` clauses that
                              must match.

    Reference: https://www.mongodb.com/docs/atlas/atlas-search/compound/
    """

    def __init__(
        self,
        *,
        must=None,
        must_not=None,
        should=None,
        filter=None,
        score=None,
        minimum_should_match=None,
    ):
        self.must = must or []
        self.must_not = must_not or []
        self.should = should or []
        self.filter = filter or []
        self.score = score
        self.minimum_should_match = minimum_should_match

    def get_search_fields(self, compiler, connection):
        fields = set()
        for clause in self.must + self.should + self.filter + self.must_not:
            fields.update(clause.get_search_fields(compiler, connection))
        return fields

    def get_source_expressions(self):
        return []

    def set_source_expressions(self, exprs):
        pass

    def resolve_expression(
        self, query=None, allow_joins=True, reuse=None, summarize=False, for_save=False
    ):
        c = self.copy()
        c.is_summary = summarize
        c.must = [
            expr.resolve_expression(query, allow_joins, reuse, summarize) for expr in self.must
        ]
        c.must_not = [
            expr.resolve_expression(query, allow_joins, reuse, summarize) for expr in self.must_not
        ]
        c.should = [
            expr.resolve_expression(query, allow_joins, reuse, summarize) for expr in self.should
        ]
        c.filter = [
            expr.resolve_expression(query, allow_joins, reuse, summarize) for expr in self.filter
        ]
        return c

    def search_operator(self, compiler, connection):
        params = {}
        if self.must:
            params["must"] = [clause.search_operator(compiler, connection) for clause in self.must]
        if self.must_not:
            params["mustNot"] = [
                clause.search_operator(compiler, connection) for clause in self.must_not
            ]
        if self.should:
            params["should"] = [
                clause.search_operator(compiler, connection) for clause in self.should
            ]
        if self.filter:
            params["filter"] = [
                clause.search_operator(compiler, connection) for clause in self.filter
            ]
        if self.minimum_should_match:
            params["minimumShouldMatch"] = self.minimum_should_match
        return {"compound": params}

    def negate(self):
        return CompoundExpression(must_not=[self])


class CombinedSearchExpression(SearchExpression):
    """
    Combine two search expressions with a logical operator such as `and`, `or`,
    or `not`.

    Example:
        CombinedSearchExpression(expr1, "and", expr2)

    Args:
        lhs: The left-hand search expression.
        operator: The boolean operator as a string (e.g., "and", "or", "not").
        rhs: The right-hand search expression.
    """

    def __init__(self, lhs, operator, rhs):
        self.lhs = lhs
        self.operator = Operator(operator) if not isinstance(operator, Operator) else operator
        self.rhs = rhs

    def get_source_expressions(self):
        return [self.lhs, self.rhs]

    def set_source_expressions(self, exprs):
        self.lhs, self.rhs = exprs

    @staticmethod
    def resolve(node, negated=False):
        if node is None:
            return None
        # Leaf, resolve the compoundExpression
        if isinstance(node, CompoundExpression):
            return node.negate() if negated else node
        # Apply De Morgan's Laws.
        operator = node.operator.negate() if negated else node.operator
        if operator == Operator.NOT:
            return node.resolve(node.lhs, not negated)
        lhs_compound = node.resolve(node.lhs, negated)
        rhs_compound = node.resolve(node.rhs, negated)
        if operator == Operator.AND:
            return CompoundExpression(must=[lhs_compound, rhs_compound])
        return CompoundExpression(should=[lhs_compound, rhs_compound], minimum_should_match=1)

    def as_mql(self, compiler, connection, as_expr=False):
        expression = self.resolve(self)
        return expression.as_mql(compiler, connection, as_expr=as_expr)


class SearchVector(SearchExpression):
    """
    Perform vector similarity search using `$vectorSearch`.

    Use the `$vectorSearch` stage to retrieve documents whose vector embeddings
    are most similar to a given query vector, according to approximate or exact
    nearest-neighbor search.

    Example:
        SearchVector("embedding", [0.1, 0.2, 0.3], limit=10, num_candidates=100)

    Args:
        path: The document path to the vector field (as string or expression).
        query_vector: The query vector to compare against.
        limit: Maximum number of matching documents to return.
        num_candidates: Optional number of candidates to consider during search.
        exact: Optional flag to enforce exact matching. The server's default is
               False, meaning approximate matching.
        filter: Optional MQL filter expression to narrow candidate documents.

    Reference: https://www.mongodb.com/docs/atlas/atlas-vector-search/vector-search-stage/
    """

    def __init__(
        self,
        path,
        query_vector,
        limit,
        *,
        num_candidates=None,
        exact=None,
        filter=None,
    ):
        self.path = cast_as_field(path)
        self.query_vector = query_vector
        self.limit = limit
        self.num_candidates = num_candidates
        self.exact = exact
        self.filter = filter
        super().__init__()

    def __invert__(self):
        raise NotSupportedError("SearchVector cannot be negated")

    def __and__(self, other):
        raise NotSupportedError("SearchVector cannot be combined")

    def __rand__(self, other):
        raise NotSupportedError("SearchVector cannot be combined")

    def __or__(self, other):
        raise NotSupportedError("SearchVector cannot be combined")

    def __ror__(self, other):
        raise NotSupportedError("SearchVector cannot be combined")

    def get_search_fields(self, compiler, connection):
        return {self.path.as_mql(compiler, connection)}

    def _get_query_index(self, fields, compiler):
        for search_indexes in compiler.collection.list_search_indexes():
            if search_indexes["type"] == "vectorSearch":
                index_field = {
                    field["path"] for field in search_indexes["latestDefinition"]["fields"]
                }
                if fields.issubset(index_field):
                    return search_indexes["name"]
        return "default"

    def as_mql(self, compiler, connection, as_expr=False):
        params = {
            "index": self._get_query_index(self.get_search_fields(compiler, connection), compiler),
            "path": self.path.as_mql(compiler, connection),
            "queryVector": self.query_vector,
            "limit": self.limit,
        }
        if self.num_candidates:
            params["numCandidates"] = self.num_candidates
        if self.exact:
            params["exact"] = self.exact
        if self.filter:
            params["filter"] = self.filter
        return {"$vectorSearch": params}


class SearchScoreOption(Expression):
    """Mutate scoring on a search operation."""

    def __init__(self, definitions=None):
        self._definitions = definitions

    def as_mql(self, compiler, connection, as_expr=False):
        return self._definitions


class SearchTextLookup(Lookup):
    """Allow QuerySet.filter(<field_name>__search="...")"""

    lookup_name = "search"

    def __init__(self, lhs, rhs):
        super().__init__(lhs, rhs)
        self.lhs = SearchText(self.lhs, self.rhs)
        self.rhs = Value(0)

    def __str__(self):
        return f"SearchText({self.lhs}, {self.rhs})"

    def __repr__(self):
        return f"SearchText({self.lhs}, {self.rhs})"

    def as_mql_expr(self, compiler, connection):
        lhs_mql = process_lhs(self, compiler, connection, as_expr=True)
        value = process_rhs(self, compiler, connection, as_expr=True)
        return {"$gte": [lhs_mql, value]}

    def as_mql_path(self, compiler, connection):
        lhs_mql = process_lhs(self, compiler, connection)
        value = process_rhs(self, compiler, connection)
        return {lhs_mql: {"$gte": value}}


CharField.register_lookup(SearchTextLookup)
TextField.register_lookup(SearchTextLookup)



import uuid
from datetime import date, datetime, time, timedelta
from decimal import Decimal

from bson import ObjectId
from django.db import DatabaseError
from django.db.models import F, Q
from django.test import SimpleTestCase

from django_mongodb_backend.fields import (
    EncryptedArrayField,
    EncryptedCharField,
    EncryptedEmbeddedModelArrayField,
    EncryptedEmbeddedModelField,
    EncryptedIntegerField,
)

from .models import (
    Actor,
    ArrayModel,
    BigIntegerModel,
    Billing,
    BinaryModel,
    BooleanModel,
    CharModel,
    DateModel,
    DateTimeModel,
    DecimalModel,
    DurationModel,
    EmailModel,
    FloatModel,
    GenericIPAddressModel,
    IntegerModel,
    Movie,
    ObjectIdModel,
    Patient,
    PatientRecord,
    PositiveBigIntegerModel,
    PositiveIntegerModel,
    PositiveSmallIntegerModel,
    SmallIntegerModel,
    TextModel,
    TimeModel,
    URLModel,
    UUIDModel,
)
from .test_base import EncryptionTestCase


class FieldTests(EncryptionTestCase):
    def assertEquality(self, model_cls, val):
        obj = model_cls.objects.create(value=val)
        self.assertEqual(model_cls.objects.get(value=val), obj)
        self.assertEqual(model_cls.objects.get(value__in=[val]), obj)
        self.assertQuerySetEqual(model_cls.objects.exclude(value=val), [])

    def assertRange(self, model_cls, *, low, high, threshold):
        obj1 = model_cls.objects.create(value=low)
        obj2 = model_cls.objects.create(value=high)
        self.assertEqual(model_cls.objects.get(value=low).value, low)
        self.assertEqual(model_cls.objects.get(value=high).value, high)
        self.assertEqual(model_cls.objects.exclude(value=high).get().value, low)
        self.assertCountEqual(model_cls.objects.filter(Q(value=high) | Q(value=low)), [obj1, obj2])
        self.assertCountEqual(model_cls.objects.filter(value__gt=threshold), [obj2])
        self.assertCountEqual(model_cls.objects.filter(value__gte=threshold), [obj2])
        self.assertCountEqual(model_cls.objects.filter(value__lt=threshold), [obj1])
        self.assertCountEqual(model_cls.objects.filter(value__lte=threshold), [obj1])
        self.assertCountEqual(model_cls.objects.filter(value__in=[low]), [obj1])
        self.assertCountEqual(model_cls.objects.filter(value__range=[low, high]), [obj1, obj2])
        msg = (
            "Comparison disallowed between Queryable Encryption encrypted "
            "fields and non-constant expressions; field 'value' is encrypted."
        )
        with self.assertRaisesMessage(DatabaseError, msg):
            model_cls.objects.filter(value__lte=F("value")).get()

    def test_array(self):
        obj = ArrayModel.objects.create(values=[1, 2, 3, 4, 5])
        obj.refresh_from_db()
        self.assertEqual(obj.values, [1, 2, 3, 4, 5])
        self.assertEncrypted(obj, "values")

    def test_embedded_model(self):
        patient = Patient.objects.create(
            patient_name="John Doe",
            patient_id=123456789,
            patient_record=PatientRecord(
                ssn="123-45-6789",
                billing=Billing(cc_type="Visa", cc_number="4111111111111111"),
            ),
        )
        patient.refresh_from_db()
        self.assertEqual(patient.patient_record.ssn, "123-45-6789")
        self.assertEqual(patient.patient_record.billing.cc_type, "Visa")
        self.assertEqual(patient.patient_record.billing.cc_number, "4111111111111111")
        self.assertEncrypted(patient, "patient_record.billing")
        self.assertEncrypted(patient, "patient_record.ssn")
        # Encrypted embedded fields are queryable.
        self.assertEqual(Patient.objects.get(patient_record__ssn="123-45-6789"), patient)

    def test_encrypted_embedded_model_field_subfields_not_queryable(self):
        """Subfields of EncryptedEmbeddedModelField aren't queryable."""
        msg = (
            "Invalid operation on path 'patient_record.billing.cc_type' which "
            "contains an encrypted path prefix."
        )
        with self.assertRaisesMessage(DatabaseError, msg):
            Patient.objects.get(patient_record__billing__cc_type="Visa")

    def test_embedded_model_array(self):
        movie = Movie.objects.create(
            title="Sample Movie",
            cast=[Actor(name="Harrison"), Actor(name="James")],
        )
        movie.refresh_from_db()
        self.assertEqual(len(movie.cast), 2)
        self.assertEqual(movie.cast[0].name, "Harrison")
        self.assertEqual(movie.cast[1].name, "James")
        self.assertEncrypted(movie, "cast")

    # Equality queries
    def test_binary(self):
        self.assertEquality(BinaryModel, b"\x00\x01\x02")
        self.assertEncrypted(BinaryModel, "value")

    def test_boolean(self):
        self.assertEquality(BooleanModel, True)
        self.assertEncrypted(BooleanModel, "value")

    def test_char(self):
        self.assertEquality(CharModel, "hello")
        self.assertEncrypted(CharModel, "value")

    def test_email(self):
        self.assertEquality(EmailModel, "test@example.com")
        self.assertEncrypted(EmailModel, "value")

    def test_generic_ip_address(self):
        self.assertEquality(GenericIPAddressModel, "192.168.0.1")
        self.assertEncrypted(GenericIPAddressModel, "value")

    def test_object_id(self):
        self.assertEquality(ObjectIdModel, ObjectId())
        self.assertEncrypted(ObjectIdModel, "value")

    def test_text(self):
        self.assertEquality(TextModel, "some text")
        self.assertEncrypted(TextModel, "value")

    def test_url(self):
        self.assertEquality(URLModel, "https://example.com")
        self.assertEncrypted(URLModel, "value")

    def test_uuid(self):
        self.assertEquality(UUIDModel, uuid.uuid4())
        self.assertEncrypted(UUIDModel, "value")

    # Range queries
    def test_big_integer(self):
        self.assertRange(BigIntegerModel, low=100, high=200, threshold=150)
        self.assertEncrypted(BigIntegerModel, "value")

    def test_date(self):
        self.assertRange(
            DateModel,
            low=date(2024, 6, 1),
            high=date(2024, 6, 10),
            threshold=date(2024, 6, 5),
        )
        self.assertEncrypted(DateModel, "value")

    def test_date_time(self):
        self.assertRange(
            DateTimeModel,
            low=datetime(2024, 6, 1, 12, 0),
            high=datetime(2024, 6, 2, 12, 0),
            threshold=datetime(2024, 6, 2, 0, 0),
        )
        self.assertEncrypted(DateTimeModel, "value")

    def test_decimal(self):
        self.assertRange(
            DecimalModel,
            low=Decimal("123.45"),
            high=Decimal("200.50"),
            threshold=Decimal("150"),
        )
        self.assertEncrypted(DecimalModel, "value")

    def test_duration(self):
        self.assertRange(
            DurationModel,
            low=timedelta(days=3),
            high=timedelta(days=10),
            threshold=timedelta(days=5),
        )
        self.assertEncrypted(DurationModel, "value")

    def test_float(self):
        self.assertRange(FloatModel, low=1.23, high=4.56, threshold=3.0)
        self.assertEncrypted(FloatModel, "value")

    def test_integer(self):
        self.assertRange(IntegerModel, low=5, high=10, threshold=7)
        self.assertEncrypted(IntegerModel, "value")

    def test_positive_big_integer(self):
        self.assertRange(PositiveBigIntegerModel, low=100, high=500, threshold=200)
        self.assertEncrypted(PositiveBigIntegerModel, "value")

    def test_positive_integer(self):
        self.assertRange(PositiveIntegerModel, low=10, high=20, threshold=15)
        self.assertEncrypted(PositiveIntegerModel, "value")

    def test_positive_small_integer(self):
        self.assertRange(PositiveSmallIntegerModel, low=5, high=8, threshold=6)
        self.assertEncrypted(PositiveSmallIntegerModel, "value")

    def test_small_integer(self):
        self.assertRange(SmallIntegerModel, low=-5, high=2, threshold=0)
        self.assertEncrypted(SmallIntegerModel, "value")

    def test_time(self):
        self.assertRange(
            TimeModel,
            low=time(10, 0),
            high=time(15, 0),
            threshold=time(12, 0),
        )
        self.assertEncrypted(TimeModel, "value")


class FieldMixinTests(SimpleTestCase):
    def test_db_index(self):
        msg = "Encrypted fields do not support db_index=True."
        with self.assertRaisesMessage(ValueError, msg):
            EncryptedIntegerField(db_index=True)

    def test_null(self):
        msg = "Encrypted fields do not support null=True."
        with self.assertRaisesMessage(ValueError, msg):
            EncryptedIntegerField(null=True)

    def test_unique(self):
        msg = "Encrypted fields do not support unique=True."
        with self.assertRaisesMessage(ValueError, msg):
            EncryptedIntegerField(unique=True)

    def test_deconstruct(self):
        field = EncryptedCharField(max_length=50, queries={"field": "value"})
        field.name = "ssn"
        name, path, args, kwargs = field.deconstruct()
        self.assertEqual(name, "ssn")
        self.assertEqual(path, "django_mongodb_backend.fields.EncryptedCharField")
        self.assertEqual(args, [])
        self.assertEqual(kwargs["queries"], {"field": "value"})

    def test_fields_without_queries(self):
        """Some field types (array, object) can't be queried."""
        for field in (
            EncryptedArrayField,
            EncryptedEmbeddedModelField,
            EncryptedEmbeddedModelArrayField,
        ):
            with self.subTest(field=field):
                msg = f"{field.__name__} does not support the queries argument."
                with self.assertRaisesMessage(ValueError, msg):
                    field(Actor, queries={})



from django.db import NotSupportedError
from django.test import SimpleTestCase

from django_mongodb_backend.expressions.search import (
    CombinedSearchExpression,
    CompoundExpression,
    SearchEquals,
    SearchVector,
)


class CombinedSearchExpressionResolutionTest(SimpleTestCase):
    def test_combined_expression_and_or_not_resolution(self):
        A = SearchEquals(path="headline", value="A")
        B = SearchEquals(path="headline", value="B")
        C = SearchEquals(path="headline", value="C")
        D = SearchEquals(path="headline", value="D")
        expr = (~A | B) & (C | D)
        solved = CombinedSearchExpression.resolve(expr)
        self.assertIsInstance(solved, CompoundExpression)
        solved_A = CompoundExpression(must_not=[CompoundExpression(must=[A])])
        solved_B = CompoundExpression(must=[B])
        solved_C = CompoundExpression(must=[C])
        solved_D = CompoundExpression(must=[D])
        self.assertCountEqual(solved.must[0].should, [solved_A, solved_B])
        self.assertEqual(solved.must[0].minimum_should_match, 1)
        self.assertEqual(solved.must[1].should, [solved_C, solved_D])

    def test_combined_expression_de_morgans_resolution(self):
        A = SearchEquals(path="headline", value="A")
        B = SearchEquals(path="headline", value="B")
        C = SearchEquals(path="headline", value="C")
        D = SearchEquals(path="headline", value="D")
        expr = ~(A | B) & (C | D)
        solved_A = CompoundExpression(must_not=[CompoundExpression(must=[A])])
        solved_B = CompoundExpression(must_not=[CompoundExpression(must=[B])])
        solved_C = CompoundExpression(must=[C])
        solved_D = CompoundExpression(must=[D])
        solved = CombinedSearchExpression.resolve(expr)
        self.assertIsInstance(solved, CompoundExpression)
        self.assertCountEqual(solved.must[0].must, [solved_A, solved_B])
        self.assertEqual(solved.must[0].minimum_should_match, None)
        self.assertEqual(solved.must[1].should, [solved_C, solved_D])
        self.assertEqual(solved.minimum_should_match, None)

    def test_combined_expression_doble_negation(self):
        A = SearchEquals(path="headline", value="A")
        expr = ~~A
        solved = CombinedSearchExpression.resolve(expr)
        solved_A = CompoundExpression(must=[A])
        self.assertIsInstance(solved, CompoundExpression)
        self.assertEqual(solved, solved_A)

    def test_combined_expression_long_right_tree(self):
        A = SearchEquals(path="headline", value="A")
        B = SearchEquals(path="headline", value="B")
        C = SearchEquals(path="headline", value="C")
        D = SearchEquals(path="headline", value="D")
        solved_A = CompoundExpression(must=[A])
        solved_B = CompoundExpression(must_not=[CompoundExpression(must=[B])])
        solved_C = CompoundExpression(must=[C])
        solved_D = CompoundExpression(must=[D])
        expr = A & ~(B & ~(C & D))
        solved = CombinedSearchExpression.resolve(expr)
        self.assertIsInstance(solved, CompoundExpression)
        self.assertEqual(len(solved.must), 2)
        self.assertEqual(solved.must[0], solved_A)
        self.assertEqual(len(solved.must[1].should), 2)
        self.assertEqual(solved.must[1].should[0], solved_B)
        self.assertCountEqual(solved.must[1].should[1].must, [solved_C, solved_D])
        expr = A | ~(B | ~(C | D))
        solved = CombinedSearchExpression.resolve(expr)
        self.assertIsInstance(solved, CompoundExpression)
        self.assertEqual(len(solved.should), 2)
        self.assertEqual(solved.should[0], solved_A)
        self.assertEqual(len(solved.should[1].must), 2)
        self.assertEqual(solved.should[1].must[0], solved_B)
        self.assertCountEqual(solved.should[1].must[1].should, [solved_C, solved_D])

    def test_vector_search_not_combinable(self):
        expr1 = SearchVector(path="headline", query_vector=[1, 2, 3], num_candidates=5, limit=2)
        expr2 = SearchVector(path="headline", query_vector=[1, 2, 4], num_candidates=5, limit=2)
        with self.assertRaisesMessage(NotSupportedError, "SearchVector cannot be combined"):
            expr1 & expr2
        with self.assertRaisesMessage(NotSupportedError, "SearchVector cannot be combined"):
            1 & expr2
        with self.assertRaisesMessage(NotSupportedError, "SearchVector cannot be combined"):
            expr1 | expr2
        with self.assertRaisesMessage(NotSupportedError, "SearchVector cannot be combined"):
            1 | expr2
        with self.assertRaisesMessage(NotSupportedError, "SearchVector cannot be negated"):
            expr1 = ~expr1



from unittest.mock import patch

from django.db import connection
from django.test import TestCase


class SupportsTransactionsTests(TestCase):
    def setUp(self):
        # Clear the cached property.
        connection.features.__dict__.pop("_supports_transactions", None)

    def tearDown(self):
        del connection.features._supports_transactions

    def test_replica_set(self):
        """A replica set supports transactions."""

        def mocked_command(command):
            if command == "hello":
                return {"setName": "foo"}
            raise Exception("Unexpected command")

        with patch("pymongo.synchronous.database.Database.command", wraps=mocked_command):
            self.assertIs(connection.features._supports_transactions, True)

    def test_sharded_cluster(self):
        """A sharded cluster supports transactions."""

        def mocked_command(command):
            if command == "hello":
                return {"msg": "isdbgrid"}
            raise Exception("Unexpected command")

        with patch("pymongo.synchronous.database.Database.command", wraps=mocked_command):
            self.assertIs(connection.features._supports_transactions, True)

    def test_no_support(self):
        """No support on a non-replica set, non-sharded cluster."""

        def mocked_command(command):
            if command == "hello":
                return {}
            raise Exception("Unexpected command")

        with patch("pymongo.synchronous.database.Database.command", wraps=mocked_command):
            self.assertIs(connection.features._supports_transactions, False)


class SupportsQueryableEncryptionTests(TestCase):
    def setUp(self):
        # Clear the cached property.
        connection.features.__dict__.pop("supports_queryable_encryption", None)
        # Must initialize cached properties before patching them.
        connection.features._supports_transactions  # noqa: B018
        connection.features.mongodb_version  # noqa: B018

    def tearDown(self):
        del connection.features.supports_queryable_encryption

    @staticmethod
    def enterprise_response(command):
        if command == "buildInfo":
            return {"modules": ["enterprise"]}
        raise Exception("Unexpected command")

    @staticmethod
    def non_enterprise_response(command):
        if command == "buildInfo":
            return {"modules": []}
        raise Exception("Unexpected command")

    def test_supported_on_atlas(self):
        """Supported on MongoDB 8.0+ Atlas replica set or sharded cluster."""
        with (
            patch(
                "pymongo.synchronous.database.Database.command", wraps=self.non_enterprise_response
            ),
            # supports_search is a proxy for "is Atlas" until MongoDB 8.2
            # when search support was added in Community Edition. Thus, this
            # may need to be reworked, however, it only affects when tests run.
            patch("django.db.connection.features.supports_search", True),
            patch("django.db.connection.features._supports_transactions", True),
            patch("django.db.connection.features.is_mongodb_8_0", True),
        ):
            self.assertIs(connection.features.supports_queryable_encryption, True)

    def test_supported_on_enterprise(self):
        """Supported on MongoDB 8.0+ Enterprise replica set or sharded cluster."""
        with (
            patch("pymongo.synchronous.database.Database.command", wraps=self.enterprise_response),
            patch("django.db.connection.features.supports_search", False),
            patch("django.db.connection.features._supports_transactions", True),
            patch("django.db.connection.features.is_mongodb_8_0", True),
        ):
            self.assertIs(connection.features.supports_queryable_encryption, True)

    def test_atlas_or_enterprise_required(self):
        """Not supported on MongoDB Community Edition."""
        with (
            patch(
                "pymongo.synchronous.database.Database.command", wraps=self.non_enterprise_response
            ),
            patch("django.db.connection.features.supports_search", False),
            patch("django.db.connection.features._supports_transactions", True),
            patch("django.db.connection.features.is_mongodb_8_0", True),
        ):
            self.assertIs(connection.features.supports_queryable_encryption, False)

    def test_transactions_required(self):
        """
        Not supported if database isn't a replica set or sharded cluster
        (i.e. DatabaseFeatures._supports_transactions = False).
        """
        with (
            patch("pymongo.synchronous.database.Database.command", wraps=self.enterprise_response),
            patch("django.db.connection.features.supports_search", False),
            patch("django.db.connection.features._supports_transactions", False),
            patch("django.db.connection.features.is_mongodb_8_0", True),
        ):
            self.assertIs(connection.features.supports_queryable_encryption, False)

    def test_mongodb_8_0_required(self):
        """Not supported on MongoDB < 8.0"""
        with (
            patch("pymongo.synchronous.database.Database.command", wraps=self.enterprise_response),
            patch("django.db.connection.features.supports_search", False),
            patch("django.db.connection.features._supports_transactions", True),
            patch("django.db.connection.features.is_mongodb_8_0", False),
        ):
            self.assertIs(connection.features.supports_queryable_encryption, False)
